import pytest
from rbc.remotejit import RemoteJIT
from rbc import rbclib
from rbc.rbclib.tracing_allocator import (TracingAllocator, LeakDetector,
                                          InvalidFreeError, MemoryLeakError)
from .test_numpy_rjit import rjit  # noqa: F401


@pytest.fixture
def djit():
    """
    Debug JIT: a RemoteJIT() which automatically uses tracing_allocator and
    detects memory leaks
    """
    return RemoteJIT(local=True, debug=True, use_tracing_allocator=True)


def test_add_ints_cffi():
    res = rbclib.lib._rbclib_add_ints(20, 22)
    assert res == 42


def test_add_ints_rjit(rjit):  # noqa: F811
    @rjit('int64(int64, int64)')
    def my_add(a, b):
        return rbclib.add_ints(a, b)
    assert my_add(20, 22) == 42


class TestTracingAllocator:

    def test_record_allocate(self):
        allocator = TracingAllocator()
        allocator.record_allocate(0x123)
        allocator.record_allocate(0x456)
        assert allocator.seq == 2
        assert allocator.alive_memory == {0x123: 1, 0x456: 2}

    def test_record_free(self):
        allocator = TracingAllocator()
        allocator.record_allocate(0x123)
        allocator.record_allocate(0x456)
        allocator.record_free(0x123)
        allocator.record_allocate(0x789)
        assert allocator.seq == 3
        assert allocator.alive_memory == {0x456: 2, 0x789: 3}

    def test_invalid_free(self):
        allocator = TracingAllocator()
        # 1. raise in case we try to free an unknown pointer
        with pytest.raises(InvalidFreeError):
            allocator.record_free(0x123)

        # 2. raise in case of double free
        allocator.record_allocate(0x456)
        allocator.record_free(0x456)
        assert allocator.alive_memory == {}
        with pytest.raises(InvalidFreeError):
            allocator.record_free(0x456)


class TestLeakDetector:

    def test_no_nested_activation(self):
        allocator = TracingAllocator()
        ld = LeakDetector(allocator)
        with ld:
            with pytest.raises(ValueError):
                with ld:
                    pass

    def test_double_enter(self):
        allocator = TracingAllocator()
        ld = LeakDetector(allocator)
        # check that we can activate the same LeakDetector twice, as long as
        # it's not nested
        with ld:
            pass
        with ld:
            pass

    def test_no_leak(self):
        allocator = TracingAllocator()
        ld = LeakDetector(allocator)
        with ld:
            allocator.record_allocate(0x123)
            allocator.record_allocate(0x456)
            allocator.record_free(0x456)
            allocator.record_free(0x123)

    def test_detect_leak(self):
        allocator = TracingAllocator()
        ld = LeakDetector(allocator)
        with pytest.raises(MemoryLeakError) as exc:
            with ld:
                allocator.record_allocate(0x123)
                allocator.record_allocate(0x456)
                allocator.record_allocate(0x789)
                allocator.record_free(0x123)
        assert exc.value.leaks == [(0x456, 2), (0x789, 3)]


class Test_djit:
    """
    These are the the most important rbclib tests: checks that we can actually
    detect memory leaks inside @djit compiled functions.
    """

    def test_djit_target_info(self, djit):
        targets = djit.targets
        ti = targets['cpu']
        assert ti.name == 'host_cpu_tracing_allocator'
        assert ti.use_tracing_allocator
        assert ti.info['fn_allocate_varlen_buffer'] == 'rbclib_tracing_allocate_varlen_buffer'

    def test_djit_leak(self, djit):
        from rbc.stdlib import array_api as xp

        @djit('int32(int32)')
        def leak_some_memory(size):
            # memory leak!
            a = xp.Array(size, xp.float64)  # noqa: F841
            return size

        with pytest.raises(MemoryLeakError):
            leak_some_memory(10)



import os
from collections import defaultdict
from rbc.omnisci_backend import Array
from rbc.errors import OmnisciServerError
from rbc.stdlib import array_api
from numba import types as nb_types
import pytest

rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def omnisci():
    # TODO: use omnisci_fixture from rbc/tests/__init__.py
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    table_name = os.path.splitext(os.path.basename(__file__))[0]

    m.sql_execute(f'DROP TABLE IF EXISTS {table_name}')
    sqltypes = ['FLOAT[]', 'DOUBLE[]',
                'TINYINT[]', 'SMALLINT[]', 'INT[]', 'BIGINT[]',
                'BOOLEAN[]']
    # todo: TEXT ENCODING DICT, TEXT ENCODING NONE, TIMESTAMP, TIME,
    # DATE, DECIMAL/NUMERIC, GEOMETRY: POINT, LINESTRING, POLYGON,
    # MULTIPOLYGON, See
    # https://www.omnisci.com/docs/latest/5_datatypes.html
    colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    m.sql_execute(f'CREATE TABLE IF NOT EXISTS {table_name} ({table_defn});')

    data = defaultdict(list)
    for i in range(5):
        for j, n in enumerate(colnames):
            if n == 'b':
                data[n].append([_i % 2 == 0 for _i in range(-3, 3)])
            elif n.startswith('f'):
                data[n].append([i * 10 + _i + 0.5 for _i in range(-3, 3)])
            else:
                data[n].append([i * 10 + _i for _i in range(-3, 3)])

    m.load_table_columnar(table_name, **data)
    m.table_name = table_name
    yield m
    try:
        m.sql_execute(f'DROP TABLE IF EXISTS {table_name}')
    except Exception as msg:
        print('%s in deardown' % (type(msg)))


@pytest.mark.parametrize('c_name', ['int8_t i1', 'int16_t i2', 'int32_t i4', 'int64_t i8',
                                    'float f4', 'double f8'])
@pytest.mark.parametrize('device', ['cpu', 'gpu'])
def test_ptr(omnisci, c_name, device):
    omnisci.reset()
    if not omnisci.has_cuda and device == 'gpu':
        pytest.skip('test requires CUDA-enabled omniscidb server')
    from rbc.external import external

    if omnisci.compiler is None:
        pytest.skip('test requires clang C/C++ compiler')

    ctype, cname = c_name.split()

    c_code = f'''
#include <stdint.h>
#ifdef __cplusplus
extern "C" {{
#endif
{ctype} mysum_impl({ctype}* x, int n) {{
    {ctype} r = 0;
    for (int i=0; i < n; i++) {{
      r += x[i];
    }}
    return r;
}}

{ctype} myval_impl({ctype}* x) {{
    return *x;
}}
#ifdef __cplusplus
}}
#endif
    '''
    omnisci.user_defined_llvm_ir[device] = omnisci.compiler(c_code)
    mysum_impl = external(f'{ctype} mysum_impl({ctype}*, int32_t)')
    myval_impl = external(f'{ctype} myval_impl({ctype}*)')

    @omnisci(f'{ctype}({ctype}[])', devices=[device])
    def mysum_ptr(x):
        return mysum_impl(x.ptr(), len(x))

    @omnisci(f'{ctype}({ctype}[], int32_t)', devices=[device])
    def myval_ptr(x, i):
        return myval_impl(x.ptr(i))

    desrc, result = omnisci.sql_execute(
        f'select {cname}, mysum_ptr({cname}) from {omnisci.table_name}')
    for a, r in result:
        if cname == 'i1':
            assert sum(a) % 256 == r % 256
        else:
            assert sum(a) == r

    desrc, result = omnisci.sql_execute(
        f'select {cname}, myval_ptr({cname}, 0), myval_ptr({cname}, 2) from {omnisci.table_name}')
    for a, r0, r2 in result:
        assert a[0] == r0
        assert a[2] == r2


def test_len_i32(omnisci):
    omnisci.reset()

    @omnisci('int64(int32[])')
    def array_sz_int32(x):
        return len(x)
    desrc, result = omnisci.sql_execute(
        f'select i4, array_sz_int32(i4) from {omnisci.table_name}')
    for a, sz in result:
        assert len(a) == sz


def test_len_f64(omnisci):
    omnisci.reset()

    @omnisci('int64(float64[])')
    def array_sz_double(x):
        return len(x)

    desrc, result = omnisci.sql_execute(
        f'select f8, array_sz_double(f8) from {omnisci.table_name}')
    for a, sz in result:
        assert len(a) == sz


@pytest.mark.skipif(available_version[:2] == (5, 1),
                    reason="skip due to a bug in omniscidb 5.1 (got %s)" % (
                        available_version,))
def test_getitem_bool(omnisci):
    omnisci.reset()

    @omnisci('bool(bool[], int64)')
    def array_getitem_bool(x, i):
        return x[i]

    query = f'select b, array_getitem_bool(b, 2) from {omnisci.table_name}'
    desrc, result = omnisci.sql_execute(query)
    for a, item in result:
        assert a[2] == item


def test_getitem_i8(omnisci):
    omnisci.reset()

    @omnisci('int8(int8[], int32)')
    def array_getitem_int8(x, i):
        return x[i]

    query = f'select i1, array_getitem_int8(i1, 2) from {omnisci.table_name}'
    desrc, result = omnisci.sql_execute(query)
    for a, item in result:
        assert a[2] == item


def test_getitem_i32(omnisci):
    omnisci.reset()

    @omnisci('int32(int32[], int32)')
    def array_getitem_int32(x, i):
        return x[i]

    query = f'select i4, array_getitem_int32(i4, 2) from {omnisci.table_name}'
    desrc, result = omnisci.sql_execute(query)
    for a, item in result:
        assert a[2] == item


def test_getitem_i64(omnisci):
    omnisci.reset()

    @omnisci('int64(int64[], int64)')
    def array_getitem_int64(x, i):
        return x[i]

    query = f'select i8, array_getitem_int64(i8, 2) from {omnisci.table_name}'
    desrc, result = omnisci.sql_execute(query)
    for a, item in result:
        assert a[2] == item


def test_getitem_float(omnisci):
    omnisci.reset()

    @omnisci('double(double[], int32)')
    def array_getitem_double(x, i):
        return x[i]

    query = f'select f8, array_getitem_double(f8, 2) from {omnisci.table_name}'
    desrc, result = omnisci.sql_execute(query)
    for a, item in result:
        assert a[2] == item
        assert type(a[2]) == type(item)

    @omnisci('float(float[], int64)')
    def array_getitem_float(x, i):
        return x[i]

    query = f'select f4, array_getitem_float(f4, 2) from {omnisci.table_name}'
    desrc, result = omnisci.sql_execute(query)
    for a, item in result:
        assert a[2] == item
        assert type(a[2]) == type(item)


def test_sum(omnisci):
    omnisci.reset()

    @omnisci('int32(int32[])')
    def array_sum_int32(x):
        r = 0
        n = len(x)
        for i in range(n):
            r = r + x[i]
        return r

    query = f'select i4, array_sum_int32(i4) from {omnisci.table_name}'
    desrc, result = omnisci.sql_execute(query)
    for a, s in result:
        assert sum(a) == s


@pytest.mark.skipif(available_version[:2] == (5, 1),
                    reason="skip due to a bug in omniscidb 5.1 (got %s)" % (
                        available_version,))
def test_even_sum(omnisci):
    omnisci.reset()

    @omnisci('int32(bool[], int32[])')
    def array_even_sum_int32(b, x):
        r = 0
        n = len(x)
        for i in range(n):
            if b[i]:
                r = r + x[i]
        return r

    query = f'select b, i4, array_even_sum_int32(b, i4) from {omnisci.table_name}'
    desrc, result = omnisci.sql_execute(query)
    for b, i4, s in result:
        assert sum([i_ for b_, i_ in zip(b, i4) if b_]) == s


def test_array_setitem(omnisci):
    omnisci.reset()

    @omnisci('double(double[], int32)')
    def array_setitem_sum(b, c):
        n = len(b)
        s = 0
        for i in range(n):
            b[i] = b[i] * c  # changes the value inplace
            s += b[i]
            b[i] = b[i] / c
        return s

    query = f'select f8, array_setitem_sum(f8, 4) from {omnisci.table_name}'
    _, result = omnisci.sql_execute(query)

    for f8, s in result:
        assert sum(f8) * 4 == s


def test_array_constructor_noreturn(omnisci):
    omnisci.reset()

    from rbc.omnisci_backend import Array

    @omnisci('float64(int32)')
    def array_noreturn(size):
        a = Array(size, nb_types.float64)
        b = Array(size, nb_types.float64)
        c = Array(size, nb_types.float64)
        for i in range(size):
            a[i] = b[i] = c[i] = i + 3.0
        s = 0.0
        for i in range(size):
            s += a[i] + b[i] + c[i] - a[i] * b[i]
        return s

    query = 'select array_noreturn(10)'
    _, result = omnisci.sql_execute(query)
    r = list(result)[0]
    assert (r == (-420.0,))


def test_array_constructor_return(omnisci):
    omnisci.reset()

    from rbc.omnisci_backend import Array
    from rbc.externals.stdio import printf

    @omnisci('float64[](int32)')
    def array_return(size):
        printf("entering array_return(%i)\n", size)
        a = Array(size, nb_types.float64)
        b = Array(size, nb_types.float64)
        for i in range(size):
            a[i] = float(i)
            b[i] = float(size - i - 1)
        if size % 2:
            c = a
        else:
            c = b
        printf("returning array with length %i\n", len(c))
        return c

    query = 'select array_return(9), array_return(10)'
    _, result = omnisci.sql_execute(query)

    r = list(result)[0]
    assert r == (list(map(float, range(9))),
                 list(map(float, reversed(range(10)))))


def test_array_constructor_len(omnisci):
    omnisci.reset()

    from rbc.omnisci_backend import Array

    @omnisci('int64(int32)')
    def array_len(size):
        a = Array(size, nb_types.float64)
        return len(a)

    query = 'select array_len(30)'
    _, result = omnisci.sql_execute(query)

    assert list(result)[0] == (30,)


def test_array_constructor_getitem(omnisci):
    omnisci.reset()

    from rbc.omnisci_backend import Array
    import numpy as np

    @omnisci('double(int32, int32)')
    def array_ptr(size, pos):
        a = Array(size, np.double)
        for i in range(size):
            a[i] = i + 0.0
        return a[pos]

    query = 'select array_ptr(5, 3)'
    _, result = omnisci.sql_execute(query)

    assert list(result)[0] == (3.0,)


def test_array_constructor_is_null(omnisci):
    omnisci.reset()

    from rbc.omnisci_backend import Array

    @omnisci('int8(int64)')
    def array_is_null(size):
        a = Array(size, 'double')
        return a.is_null()

    query = 'select array_is_null(3);'
    _, result = omnisci.sql_execute(query)

    assert list(result)[0] == (0,)


inps = [('int32', 'i4', 'trunc'), ('int32', 'i4', 'sext'),
        ('int32', 'i4', 'zext'), ('float', 'f4', 'fptrunc'),
        ('double', 'f8', 'fpext')]


@pytest.mark.parametrize("typ, col, suffix", inps,
                         ids=[item[-1] for item in inps])
def test_issue197(omnisci, typ, col, suffix):
    omnisci.reset()

    cast = dict(
        trunc=nb_types.int64,
        sext=nb_types.int8,
        zext=nb_types.uint8,
        fptrunc=nb_types.float64,
        fpext=nb_types.float32)[suffix]

    def fn_issue197(x):
        y = array_api.zeros_like(x)
        for i in range(len(x)):
            y[i] = cast(x[i] + 3)
        return y

    fn_name = f"fn_issue197_{typ}_{suffix}"
    fn_issue197.__name__ = fn_name

    omnisci(f'{typ}[]({typ}[])')(fn_issue197)

    _, result = omnisci.sql_execute(
        f'SELECT {col}, {fn_name}({col}) FROM {omnisci.table_name};'
    )

    column, ret = list(result)[0]
    for x, y in zip(column, ret):
        assert y == x + 3


def test_issue197_bool(omnisci):
    omnisci.reset()

    @omnisci('bool[](bool[])')
    def fn_issue197_bool(x):
        y = array_api.zeros_like(x)
        for i in range(len(x)):
            y[i] = bool(x[i])
        return y

    col = 'b'
    fn_name = 'fn_issue197_bool'

    _, result = omnisci.sql_execute(
        f'SELECT {col}, {fn_name}({col}) FROM {omnisci.table_name};'
    )

    column, ret = list(result)[0]
    for x, y in zip(column, ret):
        assert bool(x) == bool(y)


def test_issue109(omnisci):

    @omnisci('double[](int32)')
    def issue109(size):
        a = Array(5, 'double')
        for i in range(5):
            a[i] = nb_types.double(i)
        return a

    _, result = omnisci.sql_execute('select issue109(3);')
    assert list(result) == [([0.0, 1.0, 2.0, 3.0, 4.0],)]


def test_issue77(omnisci):

    @omnisci('int64[]()')
    def issue77():
        a = Array(5, 'int64')
        a.fill(1)
        return a

    if omnisci.version[:2] >= (5, 8):
        _, result = omnisci.sql_execute('select issue77();')
        assert list(result)[0][0] == [1, 1, 1, 1, 1]
    else:
        with pytest.raises(OmnisciServerError) as exc:
            _, result = omnisci.sql_execute('select issue77();')

        assert exc.match('Could not bind issue77()')


def test_array_dtype(omnisci):
    table = omnisci.table_name

    @omnisci('T(T[])', T=['int32', 'int64'])
    def array_dtype_fn(x):
        if x.dtype == nb_types.int32:
            return 32
        else:
            return 64

    for col, r in (('i4', 32), ('i8', 64)):
        _, result = omnisci.sql_execute(f'select array_dtype_fn({col}) from {table}')
        assert list(result) == [(r,)] * 5


def test_array_enumerate(omnisci):
    table = omnisci.table_name

    @omnisci('T(T[])', T=['int32'])
    def array_enumerate(x):
        s = 0
        for i, e in enumerate(x):
            s += e
        return s

    _, result = omnisci.sql_execute(f'select i4, array_enumerate(i4) from {table}')
    for arr, s in result:
        assert sum(arr) == s



import pytest
import re
from rbc.tests import omnisci_fixture
from rbc.errors import OmnisciServerError


@pytest.fixture(scope='module')
def omnisci():

    for o in omnisci_fixture(globals(), minimal_version=(5, 6)):
        yield o


@pytest.mark.parametrize("func", ['dbscan', 'kmeans'])
def test_mlpack(omnisci, func):
    omnisci.require_version(
        (5, 6),
        'Requires omniscidb-internal PR 5430 and omniscidb built with -DENABLE_MLPACK')

    extra_args = dict(dbscan='cast(1 as float), 1',
                      kmeans='1')[func]
    query = (f'select * from table({func}(cursor(select cast(rowid as int), f8, f8, f8 '
             f'from {omnisci.table_name}), {extra_args}, 1))')

    try:
        _, result = omnisci.sql_execute(query)
    except OmnisciServerError as msg:
        m = re.match(fr'.*Undefined function call {func!r}',
                     msg.args[0])
        if m is not None:
            pytest.skip(f'test requires omniscidb server with MLPACK support: {msg}')
        raise

    result = list(result)

    expected = dict(
        dbscan=[(0, 0), (1, 1), (2, 2), (3, 3), (4, 4)],
        kmeans=[(0, 0), (1, 0), (2, 0), (3, 0), (4, 0)]
    )[func]

    assert result == expected



"""This module implements the bi-directional conversations between
objects defined by thrift module and Python objects.

One can use class instances Buffer, NDArray as annotations in the
definitions of Dispatcher methods that are decorated with
dispatchermethod.

"""
# Author: Pearu Peterson
# Created: February 2019

import ctypes
import _ctypes
import pickle
import numpy as np


def fromobject(thrift, tcls, obj):
    """Create thrift object with type `tcls` from a Python object.
    """
    if isinstance(obj, tcls):
        return obj
    return globals().get(tcls.__name__, tcls)(thrift, obj)


def toobject(thrift, tobj, cls=None):
    """Convert thrift object `tobj` to Python object (with optional type
    `cls`).
    """
    if cls is not None and isinstance(tobj, cls):
        return tobj
    tcls = globals().get(type(tobj).__name__)
    if tcls is not None:
        return tcls.toobject(thrift, tobj, cls=cls)
    if cls is not None:
        return cls(tobj)
    return tobj


class TypeData(tuple):
    """Holds generated type for pickling:
    - constructor of the type
    - name of generated type
    - picklable data required to construct the type
    """
    @classmethod
    def fromctypes(cls, typ):
        if issubclass(typ, _ctypes.Structure):
            fields = _prepickle_dumps(tuple(typ._fields_))
            return cls((ctypes.Structure, typ.__name__, (('_fields_', fields),)))
        if issubclass(typ, _ctypes._Pointer):
            return cls((ctypes.POINTER, typ.__name__, _prepickle_dumps(typ._type_)))
        raise NotImplementedError(repr((type(typ), typ)))

    def toctypes(self):
        constructor, typname, typedata = self
        if isinstance(constructor, type) and issubclass(constructor, _ctypes.Structure):
            return type(typname, (constructor,), dict(_postpickle_loads(typedata)))
        if constructor is ctypes.POINTER:
            typ = ctypes.POINTER(_postpickle_loads(typedata))
            assert typ.__name__ == typname, (typ.__name__, typname)
            return typ
        raise NotImplementedError(repr(self))


class PointerData(tuple):
    """Holds pointer data for pickling.

    PointerData is a 4-tuple containing:
    - pointer value as int
    - value type as ctypes type
    - pointer level as int
    - sizeof value type (for integrity check)
    """

    @classmethod
    def fromctypes(cls, ptr):
        if isinstance(ptr, _ctypes._Pointer):
            level = 1   # ANSI C requires support for 12 levels as a maximum
            t = ptr._type_
            while issubclass(t, ctypes._Pointer):
                level += 1
                t = t._type_
            if t == ctypes.c_void_p:
                level += 1
            value = ctypes.cast(ptr, ctypes.c_void_p).value
            td = _prepickle_dumps(t)
            return cls((value,
                        td,
                        level,
                        ctypes.sizeof(t)))
        if isinstance(ptr, ctypes.c_void_p):
            t = type(ptr)
            value = ptr.value
            return cls((value,
                        t,
                        1,
                        ctypes.sizeof(t)))
        raise NotImplementedError(repr((type(ptr), ptr)))

    def toctypes(self):
        value, dtype, level, sizeof_dtype = self
        dtype = _postpickle_loads(dtype)
        assert ctypes.sizeof(dtype) == sizeof_dtype, (
            dtype, ctypes.sizeof(dtype), sizeof_dtype)  # TODO: mismatch of type sizes!
        if dtype == ctypes.c_void_p:
            t = dtype
        else:
            t = ctypes.POINTER(dtype)
        for i in range(1, level):
            t = ctypes.POINTER(t)
        ptr = ctypes.c_void_p(value)
        if t != ctypes.c_void_p:
            ptr = ctypes.cast(ptr, t)
        return ptr


class StructData(tuple):
    """Holds structure data for pickling.

    StructData is a 2-tuple containing:
    - structure type with _prepickle_dumps applied
    - tuple member values with _prepickle_dumps applied
    """

    @classmethod
    def fromctypes(cls, obj):
        if isinstance(obj, _ctypes.Structure):
            values = tuple([getattr(obj, name) for name, typ in obj._fields_])
            return cls(_prepickle_dumps(type(obj), values))
        raise NotImplementedError(repr((type(obj), obj)))

    def toctypes(self):
        cls, values = _postpickle_loads(self)
        return cls(*values)


def _prepickle_dumps(data):
    if isinstance(data, (_ctypes._Pointer, ctypes.c_void_p)):
        return PointerData.fromctypes(data)
    if isinstance(data, _ctypes.Structure):
        return StructData.fromctypes(data)
    if isinstance(data, tuple):
        return tuple(map(_prepickle_dumps, data))
    if isinstance(data, type) and issubclass(data, (_ctypes.Structure, _ctypes._Pointer)):
        return TypeData.fromctypes(data)
    return data


def _postpickle_loads(data):
    if isinstance(data, (PointerData, StructData, TypeData)):
        return data.toctypes()
    if isinstance(data, tuple):
        return tuple(map(_postpickle_loads, data))
    return data


class Data(object):
    """Represents any data that interpretation is defined by info and kind
    attributes.

    """
    def __new__(cls, thrift, data, info=None):
        r = thrift.Data()
        if isinstance(data, str):
            r.data = data.encode()
            r.kind = thrift.DataKind.DATA_ENCODED
            r.info = 'utf-8' if info is None else info
        elif isinstance(data, bytes):
            r.data = data
            r.kind = thrift.DataKind.DATA_RAW
            r.info = ''
        else:
            r.data = pickle.dumps(_prepickle_dumps(data))
            r.kind = thrift.DataKind.DATA_PICKLED
            r.info = str(info).encode()
        return r

    @staticmethod
    def toobject(thrift, obj, cls=None):
        kind = obj.kind
        if kind == thrift.DataKind.DATA_ENCODED:
            encoding = obj.info
            if isinstance(obj.data, str):
                return obj.data
            return obj.data.decode(encoding)
        if kind == thrift.DataKind.DATA_RAW:
            return obj.data
        if kind == thrift.DataKind.DATA_PICKLED:
            return _postpickle_loads(pickle.loads(obj.data))
        raise NotImplementedError(repr(kind))


class Buffer(object):
    """Represents a buffer that is mapped to/from Python bytes.
    """
    def __new__(cls, thrift, data):
        r = thrift.Buffer()
        if isinstance(data, thrift.Buffer):
            data = data.data
        if isinstance(data, str):
            data = data.encode()
        elif isinstance(data, np.ndarray):
            data = data.tobytes()
        if not isinstance(data, bytes):
            raise NotImplementedError('Buffer from {}'.format(type(data)))
        r.data = data
        return r

    @staticmethod
    def toobject(thrift, obj, cls=None):
        data = obj.data
        if isinstance(data, str):
            if cls is str:
                return data
            data = data.encode()
        if cls is bytes:
            return data
        if cls is np.ndarray or cls is None:
            return np.frombuffer(data, dtype=np.uint8)
        raise NotImplementedError('Buffer to {}'.format(cls))


class NDArray(object):
    """Represents N-dimensional array that is mapped to/from numpy
    ndarray.
    """

    def __new__(cls, thrift, data):
        r = thrift.NDArray()
        if isinstance(data, thrift.NDArray):
            r.data = data.data
            r.typestr = data.typestr
            r.shape = data.shape
        elif isinstance(data, np.ndarray):
            r.shape = list(data.shape)
            r.typestr = data.dtype.str
            r.data = data.tobytes()   # ensures C-contiguous array
        else:
            raise NotImplementedError('Buffer from {}'.format(type(data)))
        return r

    @staticmethod
    def toobject(thrift, obj, cls=None):
        shape = tuple(obj.shape)
        dtype = np.dtype(obj.typestr)
        data = obj.data
        if cls is np.ndarray or cls is None:
            if isinstance(data, str):
                data = data.encode()
            # C-contiguous array
            return np.frombuffer(data, dtype=dtype).reshape(shape)
        raise NotImplementedError('NDArray to {}'.format(cls))



import atexit
import pytest
import sys
import ctypes
import numpy as np
from rbc.remotejit import RemoteJIT, Signature, Caller
from rbc.typesystem import Type
from rbc.external import external
from rbc.externals.macros import sizeof, cast

win32 = sys.platform == 'win32'


@pytest.fixture(scope="module")
def rjit(request):
    local = False
    rjit = RemoteJIT(debug=not True, local=local)
    if not local:
        rjit.start_server(background=True)
        request.addfinalizer(rjit.stop_server)
        atexit.register(rjit.stop_server)
    return rjit


@pytest.fixture(scope="module")
def ljit(request):
    ljit = RemoteJIT(debug=not True, local=True)
    return ljit


def with_localjit(test_func):
    ljit = RemoteJIT(local=True)
    device = tuple(ljit.targets)[0]
    target_info = ljit.targets[device]

    def test_func_():
        with target_info:
            return test_func(ljit)
    return test_func_


def test_devices_validation(ljit):
    # normal cases

    _ = ljit('double(double, double)', devices=['cpu', 'gpu'])
    _ = ljit('double(double, double)', devices=['cpu', 'cpu', 'gpu'])

    # failing cases

    with pytest.raises(ValueError, match="'devices' can only be a list"):
        _ = ljit('double(double, double)', devices=['both'])

    with pytest.raises(ValueError, match="'devices' can only be a list"):
        _ = ljit('double(double, double)', devices=['cpu', 'gpu', 'bob'])


@with_localjit
def test_construction(ljit):

    assert isinstance(ljit, RemoteJIT)

    # Case 1

    @ljit
    def add(a: int, b: int) -> int:
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 1
    assert signatures[0] == Type.fromstring('i64(i64,i64)')

    # Case 2

    @ljit('double(double, double)')
    def add(a, b):
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 1
    assert signatures[0] == Type.fromstring('f64(f64,f64)')

    # Case 3

    @ljit('double(double, double)')
    @ljit('int(int, int)')
    def add(a, b):
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 2
    assert signatures[1] == Type.fromstring('i32(i32,i32)')
    assert signatures[0] == Type.fromstring('f64(f64,f64)')

    # Case 4

    @ljit('double(double, double)',
          'int(int, int)')
    def add(a, b):
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 2
    assert signatures[0] == Type.fromstring('f64(f64,f64)')
    assert signatures[1] == Type.fromstring('i32(i32,i32)')

    # Case 5

    ljit_int = ljit('int(int, int)')
    ljit_double = ljit('double(double, double)')
    assert isinstance(ljit_int, Signature)

    ljit_types = ljit_int(ljit_double)
    assert isinstance(ljit_types, Signature)

    @ljit_types
    def add(a, b):
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 2
    assert signatures[0] == Type.fromstring('i32(i32,i32)')
    assert signatures[1] == Type.fromstring('f64(f64,f64)')

    # Case 6

    @ljit
    def add(a, b):
        return a + b

    assert isinstance(add, Caller)
    signatures = add.get_signatures()
    assert len(signatures) == 0

    add.signature('i32(i32,i32)')
    signatures = add.get_signatures()
    assert len(signatures) == 1
    assert signatures[0] == Type.fromstring('i32(i32,i32)')

    # invalid cases are now handled at compile stage


def test_return_scalar(rjit):

    @rjit('i64(i64)', 'f64(f64)', 'c128(c128)')
    def ret(a):
        return a

    r = ret(123)
    assert r == 123
    assert isinstance(r, int)

    r = ret(123.45)
    assert r == 123.45
    assert isinstance(r, float)

    if not win32:
        # see https://github.com/xnd-project/rbc/issues/4
        r = ret(123+45j)
        assert r == 123+45j
        assert isinstance(r, complex)


def test_rjit_add(rjit):

    @rjit('i64(i64,i64)')
    def add(a, b):
        return a + b

    assert isinstance(add(1, 2), int)
    assert add(1, 2) == 3

    with pytest.raises(
            TypeError,
            match=r'found no matching function signature to given argument types'):
        add(1, 2.5)

    add.signature('d(d,d)')

    assert isinstance(add(1, 2.5), float)
    assert add(1, 2.5) == 3.5

    assert isinstance(add(1.0, 2.5), float)
    assert add(1.0, 2.5) == 3.5

    with pytest.raises(
            TypeError,
            match=r'found no matching function signature to given argument types'):
        add(1j, 2)

    add.signature('c128(c128,c128)')

    if not win32:
        # see https://github.com/xnd-project/rbc/issues/4
        r = add(1j, 2j)
        assert isinstance(r, complex)
        assert r == 3j
        assert add(1j, 2) == 2+1j


def test_options_local(rjit):

    @rjit
    def foo(x: int) -> int:
        return x + 1

    assert foo(1) == 2   # remote execution
    assert foo.local(1) == 2  # local execution


def test_composition(rjit):
    import numba as nb

    @nb.njit
    def fun(x):
        return x*x

    @rjit(local=True)
    def bar(x: int) -> int:
        y = fun(x)
        return y

    # print(bar.get_IR())


@with_localjit
def test_templates(ljit):
    ljit.reset()

    assert isinstance(ljit, RemoteJIT)

    def check_normalized(sig1, sig2):
        sig1 = set(sig1.normalized().signatures)
        sig2 = set(sig2.normalized().signatures)
        assert sig1 == sig2

    # API

    check_normalized(ljit('T(T)', templates=dict(T=['i8', 'i4'])),
                     ljit('i8(i8)', 'i4(i4)'))

    check_normalized(ljit('T(T*)', templates=dict(T=['i8', 'i4'])),
                     ljit('i8(i8*)', 'i4(i4*)'))

    check_normalized(ljit('T({T})', templates=dict(T=['i8', 'i4'])),
                     ljit('i8({i8})', 'i4({i4})'))

    check_normalized(ljit('T(T2)', templates=dict(T=['i8', 'i4'], T2=['i8', 'i4'])),
                     ljit('i8(i8)', 'i4(i4)', 'i8(i4)', 'i4(i8)'))

    check_normalized(ljit('T(T2*)', templates=dict(T=['i8', 'i4'], T2=['i8', 'i4'])),
                     ljit('i8(i8*)', 'i4(i4*)', 'i8(i4*)', 'i4(i8*)'))

    check_normalized(ljit('T({T2})', templates=dict(T=['i8', 'i4'], T2=['i8', 'i4'])),
                     ljit('i8({i8})', 'i4({i4})', 'i8({i4})', 'i4({i8})'))

    check_normalized(ljit('T({T2, T})', templates=dict(T=['i8', 'i4'], T2=['i8', 'i4'])),
                     ljit('i8({i8, i8})', 'i4({i4, i4})', 'i8({i4, i8})', 'i4({i8, i4})'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['f32'])),
                     ljit('i8(f32)', 'i4(f32)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['T'])),
                     ljit('i8(i8)', 'i4(i4)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['T*'])),
                     ljit('i8(i8*)', 'i4(i4*)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['T*', '{T*}'])),
                     ljit('i8(i8*)', 'i4(i4*)', 'i8({i8*})', 'i4({i4*})'))

    # recursion

    with pytest.raises(
            TypeError,
            match=r"cannot make T concrete"):
        ljit('T(U)', templates=dict(T=['U'], U=['T'])).normalized()

    with pytest.raises(
            TypeError,
            match=r"cannot make T concrete"):
        ljit('T(U)', templates=dict(T=['U'], U=['V'], V=['T'])).normalized()

    # not a concrete type

    with pytest.raises(
            TypeError,
            match=r"cannot make T concrete"):
        ljit('T(T)').normalized()

    # custom types

    class MyClass(Type):
        pass

    class MyClass2(Type):
        pass

    check_normalized(ljit('T(MyClass<T>)', templates=dict(T=['i8', 'i4'])),
                     ljit('i8(MyClass<int8>)', 'int4(MyClass<i4>)'))

    check_normalized(ljit('T(MyClass3<T>)', templates=dict(T=['i8', 'i4'])),
                     ljit('i8(MyClass3<int8>)', 'int4(MyClass3<i4>)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['MyClass<T>'])),
                     ljit('i8(MyClass<int8>)', 'int4(MyClass<i4>)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8', 'i4'], U=['MyClass2<T>'])),
                     ljit('i8(MyClass2<int8>)', 'int4(MyClass2<i4>)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8'], U=['C<T>'],
                                                 C=['MyClass2', 'MyClass3'])),
                     ljit('i8(MyClass2<int8>)', 'i8(MyClass3<int8>)'))

    check_normalized(ljit('T(U)', templates=dict(T=['i8'], U=['C<T>'],
                                                 C=['MyClass', 'MyClass2', 'MyClass3'])),
                     ljit('i8(MyClass2<int8>)', 'i8(MyClass<int8>)', 'i8(MyClass3<int8>)'))

    # user-friendly interface

    check_normalized(ljit('T(T)', T=['i8', 'i4']),
                     ljit('i8(i8)', 'i4(i4)'))

    check_normalized(ljit('T(U)', T=['i8', 'i4'], U=['MyClass<T>']),
                     ljit('i8(MyClass<int8>)', 'int4(MyClass<i4>)'))

    check_normalized(ljit('T(U)', T=['i8', 'i4'], U=['MyClass3<T>']),
                     ljit('i8(MyClass3<int8>)', 'int4(MyClass3<i4>)'))


scalar_pointer_types = ['int64', 'int32', 'int16', 'int8', 'float32', 'float64', 'bool', 'void']


@pytest.mark.parametrize("T", scalar_pointer_types)
@pytest.mark.parametrize("V", scalar_pointer_types)
def test_scalar_pointer_conversion(rjit, T, V):
    rjit.reset()

    with Type.alias(T=T, V=V, intp='int64'):

        # pointer-intp conversion

        @rjit('T*(intp a)')
        def i2p(a):
            return a

        @rjit('V*(intp a)')
        def i2r(a):
            return a

        @rjit('intp(T* a)')
        def p2i(a):
            return a

        @rjit('intp(V* a)')
        def r2i(a):
            return a

        @rjit('T*(T* a)')
        def p2p(a):
            return a

        @rjit('T*(V* a)')
        def p2r(a):
            return a

        @rjit('V*(T* a)')
        def r2p(a):
            return a

        # pointer artithmetics

        @rjit('T*(T* a, intp i)')
        def padd(a, i):
            return a + i

        @rjit('T*(T* a)')
        def pincr(a):
            return a + 1

        if T == V:
            i = 79
            p = i2p(i)
            assert not isinstance(p, int), (type(p), p)
            assert ctypes.cast(p, ctypes.c_void_p).value == i
            i2 = p2i(p)
            assert i2 == i
            p2 = p2p(p)
            assert ctypes.cast(p2, ctypes.c_void_p).value == ctypes.cast(p, ctypes.c_void_p).value
            assert p2i(padd(p, 2)) == i + 2
            assert p2i(pincr(p)) == i + 1

        if 'void' in [T, V]:
            i = 85
            r = i2r(i)
            assert not isinstance(r, int), (type(r), r)

            assert ctypes.cast(r, ctypes.c_void_p).value == i
            p2 = r2p(r)
            assert p2i(p2) == i
            assert p2i(p2r(i2p(i))) == i

        # double pointer-intp conversion

        @rjit('T**(intp a)')
        def i2pp(a):
            return a

        @rjit('V**(intp a)')
        def i2rr(a):
            return a

        @rjit('intp(T** a)')
        def pp2i(a):
            return a

        @rjit('intp(V** a)')
        def rr2i(a):
            return a

        @rjit('T**(T** a)')
        def pp2pp(a):
            return a

        @rjit('T**(T** a, intp i)')
        def ppadd(a, i):
            return a + i

        @rjit('T**(T** a)')
        def ppincr(a):
            return a + 1

        if T == V:
            i = 89
            pp = i2pp(i)
            assert ctypes.cast(pp, ctypes.c_void_p).value == i
            i2 = pp2i(pp)
            assert i2 == i
            pp2 = pp2pp(pp)
            assert ctypes.cast(pp2, ctypes.c_void_p).value == ctypes.cast(
                pp, ctypes.c_void_p).value
            assert pp2i(ppadd(pp, 2)) == i + 2
            assert pp2i(ppincr(pp)) == i + 1

        # double pointer-pointer conversion

        @rjit('T**(T* a)')
        def p2pp(a):
            return a

        @rjit('T*(T** a)')
        def pp2p(a):
            return a

        # mixed pointer conversion requires void*

        @rjit('T**(V* a)')
        def r2pp(a):
            return a

        @rjit('V*(T** a)')
        def pp2r(a):
            return a

        if T == V == 'void':
            i = 99
            p = i2p(i)
            pp = p2pp(p)
            assert pp2i(pp) == i
            assert p2i(pp2p(pp)) == i

        if V == 'void':
            i = 105
            assert r2i(pp2r(i2pp(i))) == i
            assert pp2i(r2pp(i2r(i))) == i


@pytest.mark.parametrize("T", ['int64', 'int32', 'int16', 'int8', 'float32', 'float64'])
def test_scalar_pointer_access_local(ljit, T):
    ljit.reset()

    with Type.alias(T=T):
        arr = np.array([1, 2, 3, 4], dtype=T)
        arr2 = np.array([11, 12, 13, 14], dtype=T)

        ptr = ctypes.c_void_p(arr.__array_interface__['data'][0])

        @ljit('T(T* x, int i)')
        def pitem(x, i):
            return x[i]

        @ljit('void(T* x, int i, T)')
        def sitem(x, i, v):
            x[i] = v

        for i in range(len(arr)):
            v = pitem(ptr, i)
            assert v == arr[i]

        for i in range(len(arr)):
            sitem(ptr, i, arr2[i])

        assert (arr == arr2).all()


memory_managers = dict(
    cstdlib=('void* calloc(size_t nmemb, size_t size)',
             'void free(void* ptr)'),
    PyMem_Raw=(
        'void* PyMem_RawCalloc(size_t nmemb, size_t size)',
        'void PyMem_RawFree(void* ptr)'),
    PyMem=(
        'void* PyMem_Calloc(size_t nmemb, size_t size)',
        'void PyMem_Free(void* ptr)'),
)


@pytest.mark.parametrize("T", ['int64', 'int32', 'int16', 'int8', 'float32', 'float64'])
@pytest.mark.parametrize("memman", list(memory_managers))
def test_scalar_pointer_access_remote(rjit, memman, T):
    rjit.reset()

    calloc_prototype, free_prototype = memory_managers[memman]

    with Type.alias(T=T):

        arr = np.array([1, 2, 3, 4], dtype=T)
        arr2 = np.array([11, 12, 13, 14], dtype=T)

        calloc = external(calloc_prototype)
        free = external(free_prototype)

        @rjit(calloc_prototype)
        def rcalloc(nmemb, size):
            return calloc(nmemb, size)

        @rjit(free_prototype)
        def rfree(ptr):
            return free(ptr)

        @rjit('T(T* x, int i)')
        def pitem(x, i):
            return x[i]

        @rjit('void(T* x, int i, T)')
        def sitem(x, i, v):
            x[i] = v

        ptr = rcalloc(arr.itemsize, len(arr))

        assert ptr.value

        for i in range(len(arr)):
            sitem(ptr, i, arr[i])

        for i in range(len(arr2)):
            arr2[i] = pitem(ptr, i)

        rfree(ptr)

        assert (arr == arr2).all()


@pytest.mark.parametrize("location", ['local', 'remote'])
@pytest.mark.parametrize("T", ['int64', 'int32', 'int16', 'int8', 'float32', 'float64'])
def test_struct_input(ljit, rjit, location, T):
    jit = rjit if location == 'remote' else ljit
    jit.reset()

    S = '{T x, T y, T z}'

    with Type.alias(T=T, S=S):

        nb_S = Type.fromstring(S).tonumba()
        nb_T = Type.fromstring(T).tonumba()

        class MetaMyStructType(type):

            @property
            def __typesystem_type__(cls):
                # Python type dependent type
                return Type.fromstring(S) | type(cls).__name__

        class MyStruct(tuple, metaclass=MetaMyStructType):

            @property
            def __typesystem_type__(self):
                # Python instance dependent Type
                return Type.fromstring(S) | type(self).__name__

            @classmethod
            def fromobject(cls, obj):
                return cls([getattr(obj, m.name) for m in cls.__typesystem_type__])

            def __getattr__(self, name):
                typ = self.__typesystem_type__
                index = typ.get_field_position(name)
                return self[index]

        x, y, z = 123, 2, 3
        s = MyStruct((x, y, z))

        assert type(s).__typesystem_type__ == s.__typesystem_type__
        assert type(s).__typesystem_type__.annotation() != s.__typesystem_type__.annotation()

        # Get the members of struct given by value and reference

        @jit('T(S)', 'T(S*)')
        def get_x(s): return s.x

        @jit('T(S)', 'T(S*)')
        def get_y(s): return s.y

        @jit('T(S)', 'T(S*)')
        def get_z(s): return s.z

        assert get_x(s) == x
        assert get_y(s) == y
        assert get_z(s) == z

        # Set the members of struct given by reference
        # Manage memory of the arrays with struct values

        rcalloc, rfree = map(external, memory_managers['cstdlib'])

        @jit('S* allocate_S(size_t i)')
        def allocate_S(i):
            return rcalloc(i, sizeof(nb_S))

        @jit('void free(void*)')
        def free(s):
            rfree(s)

        @jit('void set_x(S*, T)')
        def set_x(s, x):
            s.x = x

        @jit('void set_y(S*, T)')
        def set_y(s, y):
            s.y = y

        @jit('void set_z(S*, T)')
        def set_z(s, z):
            s.z = z

        p = allocate_S(1)

        set_x(p, x)
        assert get_x(p) == x

        set_y(p, y)
        assert get_y(p) == y

        set_z(p, z)
        assert get_z(p) == z

        free(p)

        @jit('size_t sizeof_S(S)', 'size_t sizeof_S(S*)', 'size_t sizeof_T(T)')
        def sizeof_S(s):
            return sizeof(s)

        assert sizeof_S(s) == sizeof_S(nb_T(0)) * 3
        assert sizeof_S(p) == 8

        # Pointer arithmetics and getitem

        @jit('void(S*, int64)')
        def init_S(s, n):
            s.x = 8
            for i in range(n):
                si = s + i  # pointer arithmetics
                si.x = i
                si.y = i*10
                si.z = i*20

        @jit('void(S*, int64)')
        def init_S2(s, n):
            s.x = 8
            for i in range(n):
                # `s[i]` is equivalent to `s + i`! That is, the result
                # of `s[i]` is a pointer value `S*` rather than a
                # value of `S`.
                si = s[i]
                si.x = i*5
                si.y = i*15
                si.z = i*25

        @jit('S*(S*)')
        def next_S(s):
            return s + 1

        n = 5
        p = allocate_S(n)
        init_S(p, n)

        r = p
        for i in range(n):
            assert (get_x(r), get_y(r), get_z(r)) == (i, i*10, i*20)
            r = next_S(r)

        init_S2(p, n)

        r = p
        for i in range(n):
            assert (get_x(r), get_y(r), get_z(r)) == (i*5, i*15, i*25)
            r = next_S(r)

        free(p)


@pytest.mark.parametrize("location", ['local', 'remote'])
@pytest.mark.parametrize("T", ['int64', 'int32', 'int16', 'int8', 'float32', 'float64'])
def test_struct_pointer_members(ljit, rjit, location, T):
    jit = rjit if location == 'remote' else ljit
    jit.reset()

    index_t = 'int32'
    S = '{T* data, index_t size}'

    with Type.alias(T=T, S=S, index_t=index_t):
        Sp = S + '*'
        nb_S = Type.fromstring(S).tonumba()
        nb_Sp = Type.fromstring(Sp).tonumba()
        nb_T = Type.fromstring(T).tonumba()

        rcalloc, rfree = map(external, memory_managers['cstdlib'])

        @jit('S* allocate_S(size_t i)')
        def allocate_S(i):
            raw = rcalloc(i, sizeof(nb_S))
            s = cast(raw, nb_Sp)     # cast `void*` to `S*`
            # s = raw.cast(Sp)       # cast `void*` to `S*`
            s.size = 0
            return s

        @jit('void free(void*)')
        def free(s):
            rfree(s)

        @jit('index_t getsize(S*)')
        def getsize(s):
            return s.size

        @jit('T getitem(S*, index_t)')
        def getitem(s, i):
            if i >= 0 and i < s.size:
                return s.data[i]
            return 0  # TODO: return some null value or raise exception

        @jit('void resize_S(S*, index_t)')
        def resize_S(s, size):
            if s.size == size:
                return
            if s.size != 0:
                rfree(s.data)
                s.size = 0
            if s.size == 0 and size > 0:
                raw = rcalloc(size, sizeof(nb_T))
                s.data = raw
                s.size = size

        s = allocate_S(1)
        assert getsize(s) == 0
        assert getitem(s, 0) == 0

        n = 10
        resize_S(s, n)
        assert getsize(s) == n

        n = 5
        resize_S(s, n)
        assert getsize(s) == n

        @jit('void fill_S(S*, T)')
        def fill_S(s, v):
            for i in range(s.size):
                s.data[i] = v

        fill_S(s, 123)
        assert getitem(s, 0) == 123

        @jit('void range_S(S*)')
        def range_S(s):
            for i in range(s.size):
                s.data[i] = i

        range_S(s)

        for i in range(getsize(s)):
            assert getitem(s, i) == i

        resize_S(s, 0)  # deallocate data
        free(s)         # deallocate struct


@pytest.mark.parametrize("location", ['local', 'remote'])
@pytest.mark.parametrize("T", ['int64', 'int32', 'int16', 'int8', 'float32', 'float64'])
def test_struct_double_pointer_members(ljit, rjit, location, T):
    jit = rjit if location == 'remote' else ljit
    jit.reset()

    index_t = 'int32'
    S = '{T** data, index_t length, index_t size}'
    S1 = '{T* data, index_t size}'

    with Type.alias(T=T, S=S, S1=S1, index_t=index_t):
        Sp = S + '*'
        S1p = S1 + '*'
        Tp = T + '*'
        Tpp = T + '**'
        nb_S = Type.fromstring(S).tonumba()
        nb_Sp = Type.fromstring(Sp).tonumba()
        nb_S1 = Type.fromstring(S1).tonumba()
        nb_S1p = Type.fromstring(S1p).tonumba()
        nb_T = Type.fromstring(T).tonumba()
        nb_Tp = Type.fromstring(Tp).tonumba()
        nb_Tpp = Type.fromstring(Tpp).tonumba()

        rcalloc, rfree = map(external, memory_managers['cstdlib'])

        @jit('S* allocate_S(size_t length, size_t size)')
        def allocate_S(length, size):
            s = rcalloc(1, sizeof(nb_S)).cast(nb_Sp)
            s.data = rcalloc(length, sizeof(nb_Tp)).cast(nb_Tpp)
            s.length = length
            for i in range(length):
                s.data[i] = rcalloc(size, sizeof(nb_T)).cast(nb_Tp)
            s.size = size
            return s

        @jit('void deallocate_S(S*)')
        def deallocate_S(s):
            if s.length > 0:
                for i in range(s.length):
                    rfree(s.data[i])
                rfree(s.data)
                rfree(s)

        @jit('S1* viewitem(S*, index_t)')
        def viewitem(s, i):
            s1 = rcalloc(1, sizeof(nb_S1)).cast(nb_S1p)
            if i >= 0 and i < s.size:
                s1.size = s.size
                s1.data = s.data[i]
            else:
                s1.size = 0
            return s1

        @jit('T getitem(S*, index_t, index_t)')
        def getitem(s, col, row):
            if col >= 0 and col < s.length:
                if row >= 0 and row < s.size:
                    return s.data[col][row]
            return 0

        @jit('void free(S1*)')
        def free(s1):
            rfree(s1)

        @jit('void fill1(S1*, T)')
        def fill1(s1, v):
            for i in range(s1.size):
                s1.data[i] = v

        @jit('void range1(S1*)')
        def range1(s1):
            for i in range(s1.size):
                s1.data[i] = i

        m, n = 4, 5
        s = allocate_S(m, n)

        # initialize
        expected = []
        for col in range(m):
            s1 = viewitem(s, col)
            if col % 2:
                expected.extend(n * [23+col*10])
                fill1(s1, 23+col*10)
            else:
                expected.extend(range(n))
                range1(s1)
            free(s1)

        # check
        actual = []
        for col in range(m):
            for row in range(n):
                actual.append(getitem(s, col, row))
        deallocate_S(s)

        assert expected == actual



import pytest
from rbc.tests import omnisci_fixture
from rbc.external import external
from numba import types


@pytest.fixture(scope="module")
def omnisci():
    for o in omnisci_fixture(globals()):
        yield o


def test_valid_signatures(omnisci):
    external("f64 log2(f64)")
    assert True


def test_invalid_signature(omnisci):
    with pytest.raises(ValueError) as excinfo:
        external(types.int64)

    assert "signature must represent a function type" in str(excinfo)


def test_unnamed_external(omnisci):
    with pytest.raises(ValueError) as excinfo:
        external("f64(f64)")

    assert "external function name not specified for signature" in str(excinfo)

    with pytest.raises(ValueError) as excinfo:
        external("f64(f64)")

    assert "external function name not specified for signature" in str(excinfo)


def test_replace_declaration(omnisci):

    _ = external("f64 fma(f64)|CPU")
    fma = external("f64 fma(f64, f64, f64)|CPU")

    @omnisci("double(double, double, double)", devices=["cpu"])
    def test_fma(a, b, c):
        return fma(a, b, c)

    _, result = omnisci.sql_execute("select test_fma(1.0, 2.0, 3.0)")

    assert list(result) == [(5.0,)]


def test_require_target_info(omnisci):

    log2 = external("double log2(double)|CPU")

    @omnisci("double(double)", devices=["cpu"])
    def test_log2(a):
        return log2(a)

    _, result = omnisci.sql_execute("select test_log2(8.0)")

    assert list(result) == [(3.0,)]



import os
from collections import defaultdict
import pytest

rbc_omnisci = pytest.importorskip('rbc.omniscidb')
available_version, reason = rbc_omnisci.is_available()
pytestmark = pytest.mark.skipif(not available_version, reason=reason)


@pytest.fixture(scope='module')
def omnisci():
    # TODO: use omnisci_fixture from rbc/tests/__init__.py
    config = rbc_omnisci.get_client_config(debug=not True)
    m = rbc_omnisci.RemoteOmnisci(**config)
    table_name = os.path.splitext(os.path.basename(__file__))[0]

    m.sql_execute(f'DROP TABLE IF EXISTS {table_name}')

    sqltypes = ['TEXT ENCODING DICT', 'TEXT ENCODING DICT(16)', 'TEXT ENCODING DICT(8)',
                'TEXT[] ENCODING DICT(32)', 'TEXT ENCODING NONE', 'TEXT ENCODING NONE', 'INT[]']
    colnames = ['t4', 't2', 't1', 's', 'n', 'n2', 'i4']
    table_defn = ',\n'.join('%s %s' % (n, t)
                            for t, n in zip(sqltypes, colnames))
    m.sql_execute(f'CREATE TABLE IF NOT EXISTS {table_name} ({table_defn});')

    data = defaultdict(list)
    for i in range(5):
        for j, n in enumerate(colnames):
            # todo: to check is_null, use empty string
            if n in ['t4', 't2']:
                data[n].append(['foofoo', 'bar', 'fun', 'bar', 'foo'][i])
            if n in ['t1', 'n']:
                data[n].append(['fun', 'bar', 'foo', 'barr', 'foooo'][i])
            elif n == 's':
                data[n].append([['foo', 'bar'], ['fun', 'bar'], ['foo']][i % 3])
            elif n == 'n2':
                data[n].append(['1', '12', '123', '1234', '12345'][i])
            elif n == 'i4':
                data[n].append(list(range(i+1)))

    m.load_table_columnar(table_name, **data)
    m.table_name = table_name
    yield m
    try:
        m.sql_execute(f'DROP TABLE IF EXISTS {table_name}')
    except Exception as msg:
        print('%s in deardown' % (type(msg)))


def test_table_data(omnisci):
    omnisci.reset()

    descr, result = omnisci.sql_execute(
        f'select t4, t2, t1, s, n from {omnisci.table_name}')
    result = list(result)

    assert result == [('foofoo', 'foofoo', 'fun', ['foo', 'bar'], 'fun'),
                      ('bar', 'bar', 'bar', ['fun', 'bar'], 'bar'),
                      ('fun', 'fun', 'foo', ['foo'], 'foo'),
                      ('bar', 'bar', 'barr', ['foo', 'bar'], 'barr'),
                      ('foo', 'foo', 'foooo', ['fun', 'bar'], 'foooo')]


def test_bytes_len(omnisci):
    omnisci.reset()

    @omnisci('int32(Bytes, Bytes)')
    def mystrlen(s, s2):
        return len(s) * 100 + len(s2)

    sql_query_expected = f"select length(n) * 100 + length(n2) from {omnisci.table_name}"
    sql_query = f"select mystrlen(n, n2) from {omnisci.table_name}"

    descr, result_expected = omnisci.sql_execute(sql_query_expected)
    result_expected = list(result_expected)
    descr, result = omnisci.sql_execute(sql_query)
    result = list(result)

    assert result == result_expected


def test_bytes_ord(omnisci):
    omnisci.reset()

    @omnisci('int64(Bytes, int32)', devices=['gpu', 'cpu'])
    def myord(s, i):
        return s[i] if i < len(s) else 0

    sql_query_data = f"select n from {omnisci.table_name}"
    descr, data = omnisci.sql_execute(sql_query_data)
    data = list(data)

    for i in range(5):
        result_expected = [(ord(d[0][i]) if i < len(d[0]) else 0, ) for d in data]
        sql_query = f"select myord(n, {i}) from {omnisci.table_name}"
        descr, result = omnisci.sql_execute(sql_query)
        result = list(result)
        assert result == result_expected


def test_bytes_return(omnisci):
    omnisci.reset()

    from rbc.omnisci_backend import Bytes

    @omnisci('Bytes(int32, int32)')
    def make_abc(first, n):
        r = Bytes(n)
        for i in range(n):
            r[i] = first + i
        return r

    sql_query = "select make_abc(97, 10)"
    descr, result = omnisci.sql_execute(sql_query)
    result = list(result)

    assert result == [('abcdefghij', )]


def test_bytes_upper(omnisci):
    omnisci.reset()

    from rbc.omnisci_backend import Bytes

    @omnisci('Bytes(Bytes)')
    def myupper(s):
        r = Bytes(len(s))
        for i in range(len(s)):
            c = s[i]
            if c >= 97 and c <= 122:
                c = c - 32
            r[i] = c
        return r

    sql_query = f"select n, myupper(n) from {omnisci.table_name}"
    descr, result = omnisci.sql_execute(sql_query)
    result = list(result)

    for n, u in result:
        assert n.upper() == u



'''Omnisci Bytes type that corresponds to Omnisci type TEXT ENCODED NONE.

Omnisci Bytes represents the following structure:

  struct Bytes {
    char* ptr;
    size_t sz;  // when non-negative, Bytes has fixed width.
  }
'''

__all__ = ['BytesPointer', 'Bytes', 'OmnisciBytesType']

from rbc import typesystem
from .omnisci_buffer import (
    BufferPointer, Buffer, OmnisciBufferType,
    omnisci_buffer_constructor)
from numba.core import types, extending


class OmnisciBytesType(OmnisciBufferType):
    """Omnisci Bytes type for RBC typesystem.
    """

    @property
    def buffer_extra_members(self):
        return ('bool is_null',)

    def match(self, other):
        if type(self) is type(other):
            return self[0] == other[0]
        if other.is_pointer and other[0].is_char and other[0].bits == 8:
            return 1
        if other.is_string:
            return 2


BytesPointer = BufferPointer


class Bytes(Buffer):
    pass


@extending.lower_builtin(Bytes, types.Integer)
def omnisci_bytes_constructor(context, builder, sig, args):
    return omnisci_buffer_constructor(context, builder, sig, args)


@extending.type_callable(Bytes)
def type_omnisci_bytes(context):
    def typer(size):
        return typesystem.Type.fromobject('Bytes').tonumba()
    return typer



from rbc.stdlib import array_api as xp
from rbc.omnisci_backend.omnisci_buffer import free_buffer
from ..test_rbclib import djit  # noqa: F401


def test_array_free_function_call(djit):    # noqa: F811

    @djit('int32(int32)')
    def fn(size):
        a = xp.Array(size, xp.float64)
        free_buffer(a)
        return size

    res = fn(10)
    assert res == 10


def test_array_free_method(djit):    # noqa: F811

    @djit('int32(int32)')
    def fn(size):
        a = xp.Array(size, xp.float64)
        a.free()
        return size

    res = fn(10)
    assert res == 10


def test_array_constructor_noreturn(djit):    # noqa: F811

    @djit('float64(int32)')
    def array_noreturn(size):
        a = xp.Array(size, xp.float64)
        b = xp.Array(size, xp.float64)
        c = xp.Array(size, xp.float64)
        for i in range(size):
            a[i] = b[i] = c[i] = i + 3.0
        s = 0.0
        for i in range(size):
            s += a[i] + b[i] + c[i] - a[i] * b[i]
        a.free()
        b.free()
        c.free()
        return s

    res = array_noreturn(10)
    assert res == -420



import numpy as np
import pytest
from rbc.externals.omniscidb import set_output_row_size
from rbc.omnisci_backend import Bytes
from rbc.tests import omnisci_fixture, assert_equal


@pytest.fixture(scope='module')
def omnisci():
    for o in omnisci_fixture(globals(), minimal_version=(5, 8)):
        define(o)
        yield o


def define(omnisci):

    @omnisci('T(T)', T=['int32', 'int64', 'float32', 'float64'])
    def myincr(x):
        return x + 1

    T = ['int64', 'float64', 'int32']

    @omnisci('UDTF(int32 size, T x0, OutputColumn<T> x)', T=T, devices=['cpu'])
    def arange(size, x0, x):
        set_output_row_size(size)
        for i in range(size):
            x[i] = x0 + x.dtype(i)
        return size

    T = ['int64', 'float32']

    @omnisci('UDTF(Column<T> x, T dx, OutputColumn<T> y)', T=T, devices=['cpu'])
    def aincr(x, dx, y):
        size = len(x)
        set_output_row_size(size)
        for i in range(size):
            y[i] = x[i] + dx
        return size

    @omnisci('Bytes(Bytes)')
    def myupper(s):
        r = Bytes(len(s))
        for i in range(len(s)):
            c = s[i]
            if c >= 97 and c <= 122:
                c = c - 32
            r[i] = c
        return r


def test_udf_string_repr(omnisci):
    myincr = omnisci.get_caller('myincr')

    assert_equal(repr(myincr),
                 "RemoteDispatcher('myincr', ['T(T), T=int32|int64|float32|float64'])")
    assert_equal(str(myincr), "myincr['T(T), T=int32|int64|float32|float64']")

    assert_equal(repr(myincr(5)),
                 "OmnisciQueryCapsule('SELECT myincr(CAST(5 AS BIGINT))')")
    assert_equal(str(myincr(5)), "SELECT myincr(CAST(5 AS BIGINT))")

    assert_equal(repr(myincr(myincr(5))),
                 "OmnisciQueryCapsule('SELECT myincr(CAST(myincr(CAST(5 AS BIGINT)) AS BIGINT))')")
    assert_equal(str(myincr(myincr(5))),
                 "SELECT myincr(CAST(myincr(CAST(5 AS BIGINT)) AS BIGINT))")


def test_udtf_string_repr(omnisci):
    arange = omnisci.get_caller('arange')

    assert_equal(repr(arange),
                 ("RemoteDispatcher('arange', ['UDTF(int32 size, T x0, OutputColumn<T> x),"
                  " T=int64|float64|int32, device=cpu'])"))
    assert_equal(str(arange),
                 ("arange['UDTF(int32 size, T x0, OutputColumn<T> x),"
                  " T=int64|float64|int32, device=cpu']"))

    assert_equal(repr(arange(5, 0)),
                 ("OmnisciQueryCapsule('SELECT x FROM"
                  " TABLE(arange(CAST(5 AS INT), CAST(0 AS BIGINT)))')"))
    assert_equal(str(arange(5, 0)),
                 "SELECT x FROM TABLE(arange(CAST(5 AS INT), CAST(0 AS BIGINT)))")


def test_remote_udf_evaluation(omnisci):
    myincr = omnisci.get_caller('myincr')

    assert_equal(str(myincr(3)), 'SELECT myincr(CAST(3 AS BIGINT))')
    assert_equal(myincr(3, hold=False), 4)
    assert_equal(myincr(3).execute(), 4)

    assert_equal(myincr(3.5).execute(), 4.5)
    assert_equal(myincr(np.int64(3)).execute(), np.int64(4))
    assert_equal(myincr(np.float32(3.5)).execute(), np.float32(4.5))


def test_remote_int32_evaluation(omnisci):
    myincr = omnisci.get_caller('myincr')
    arange = omnisci.get_caller('arange')

    pytest.xfail('SELECT upcasts int32 to int64')
    assert_equal(myincr(np.int32(3)), np.int32(4))
    assert_equal(arange(3, np.int32(1))['x'], np.arange(3, dtype=np.int32) + 1)


def test_remote_float64_evaluation(omnisci):
    myincr = omnisci.get_caller('myincr')
    arange = omnisci.get_caller('arange')

    pytest.xfail('SELECT downcasts float64 to float32')
    assert_equal(myincr(np.float64(3.5)), np.float64(4.5))
    assert_equal(arange(3, np.float64(1))['x'], np.arange(3, dtype=np.float64) + 1)


def test_remote_bytes_evaluation(omnisci):
    myupper = omnisci.get_caller('myupper')
    assert str(myupper) == "myupper['Bytes(Bytes)']"
    assert str(myupper("abc")) == "SELECT myupper('abc')"
    assert str(myupper("abc").execute()) == 'ABC'
    assert str(myupper(b"abc")) == "SELECT myupper('abc')"
    assert str(myupper(b"abc").execute()) == 'ABC'


def test_remote_composite_udf_evaluation(omnisci):
    myincr = omnisci.get_caller('myincr')

    assert_equal(str(myincr(myincr(3))),
                 'SELECT myincr(CAST(myincr(CAST(3 AS BIGINT)) AS BIGINT))')
    assert_equal(str(myincr(myincr(3, hold=False))), 'SELECT myincr(CAST(4 AS BIGINT))')
    assert_equal(myincr(myincr(3), hold=False), 5)
    assert_equal(myincr(myincr(3)).execute(), 5)


def test_remote_udtf_evaluation(omnisci):
    arange = omnisci.get_caller('arange')

    assert_equal(str(arange(3, 1)),
                 'SELECT x FROM TABLE(arange(CAST(3 AS INT), CAST(1 AS BIGINT)))')

    assert_equal(arange(3, 1).execute()['x'], list(np.arange(3, dtype=np.int64) + 1))
    assert_equal(arange(3, 1.5).execute()['x'], list(np.arange(3, dtype=np.float64) + 1.5))
    assert_equal(arange(np.int32(3), 1).execute()['x'], list(np.arange(3, dtype=np.int32) + 1))
    assert_equal(arange(3, np.float32(1)).execute()['x'], np.arange(3, dtype=np.float32) + 1)
    assert_equal(arange(3, np.int64(1)).execute()['x'], np.arange(3, dtype=np.int64) + 1)


def test_remote_composite_udtf_evaluation(omnisci):
    arange = omnisci.get_caller('arange')
    aincr = omnisci.get_caller('aincr')
    myincr = omnisci.get_caller('myincr')

    r = aincr(arange(3, 1), 2)
    assert_equal(str(r), 'SELECT y FROM TABLE(aincr(CURSOR(SELECT x FROM'
                 ' TABLE(arange(CAST(3 AS INT), CAST(1 AS BIGINT)))), CAST(2 AS BIGINT)))')

    r = r.execute()
    assert_equal(r['y'], np.arange(3, dtype=np.int64) + 1 + 2)

    r = arange(3, myincr(2, hold=False))
    assert_equal(str(r), 'SELECT x FROM TABLE(arange(CAST(3 AS INT), CAST(3 AS BIGINT)))')
    assert_equal(r.execute()['x'], np.arange(3, dtype=np.int64) + 2 + 1)


def test_remote_composite_udtf_udf(omnisci):
    """
    TableFunctionExecutionContext.cpp:277 Check failed:
    col_buf_ptrs.size() == exe_unit.input_exprs.size() (1 == 2)
    """
    myincr = omnisci.get_caller('myincr')
    arange = omnisci.get_caller('arange')

    r = arange(3, myincr(2))
    assert_equal(str(r), ('SELECT x FROM TABLE(arange(CAST(3 AS INT),'
                          ' CAST(myincr(CAST(2 AS BIGINT)) AS BIGINT)))'))

    pytest.xfail('udtf(udf) crashes omniscidb server')
    assert_equal(r['x'], np.arange(3, dtype=np.int64) + 2 + 1)


def test_remote_udf_typeerror(omnisci):
    myincr = omnisci.get_caller('myincr')
    try:
        myincr("abc")
    except TypeError as msg:
        assert_equal(str(msg), '''\
found no matching function signature to given argument types:
    (string) -> ...
  available function signatures:
    (int32) -> int32
    (int64) -> int64
    (float32) -> float32
    (float64) -> float64''')
    else:
        assert 0  # expected TypeError


def test_remote_udtf_typeerror(omnisci):
    arange = omnisci.get_caller('arange')
    try:
        arange(1.2, 0)
    except TypeError as msg:
        assert_equal(str(msg), '''\
found no matching function signature to given argument types:
    (float64, int64) -> ...
  available function signatures:
    (int32 size, int64 x0) -> (Column<int64> x)
    - UDTF(int32 size, int64 x0, OutputColumn<int64> x)
    (int32 size, float64 x0) -> (Column<float64> x)
    - UDTF(int32 size, float64 x0, OutputColumn<float64> x)
    (int32 size, int32 x0) -> (Column<int32> x)
    - UDTF(int32 size, int32 x0, OutputColumn<int32> x)''')
    else:
        assert 0  # expected TypeError


def test_remote_udf_overload(omnisci):

    @omnisci('int32(int32)')  # noqa: F811
    def incr_ol(x):           # noqa: F811
        return x + 1

    @omnisci('int32(int32, int32)')  # noqa: F811
    def incr_ol(x, dx):              # noqa: F811
        return x + dx

    assert incr_ol(1).execute() == 2
    assert incr_ol(1, 2).execute() == 3



from ._rbclib import lib, ffi
from .errors import TracingAllocatorError, InvalidFreeError, MemoryLeakError  # noqa: F401


class TracingAllocator:
    """
    Provide debug versions of allocate_varlen_buffer and free_buffer which
    keep trace of all the allocation/deallocations in order to detect memory
    leaks.

    The logic is written in pure Python, and it is exposed to the C world
    through CFFI's @def_extern() mechanism.
    """

    def __init__(self):
        # alive_memory is a dictionary which contains all the addresses which
        # have been allocated but not yet freed.  For each address we record
        # an unique sequence number which acts as a timestamp, so that we can
        # inspect them in allocation order.
        self.seq = 0
        self.alive_memory = {}  # {address: seq}

    def record_allocate(self, addr):
        self.seq += 1
        assert addr not in self.alive_memory
        self.alive_memory[addr] = self.seq

    def record_free(self, addr):
        if addr not in self.alive_memory:
            raise InvalidFreeError('Trying to free() a dangling pointer?')
        del self.alive_memory[addr]


class LeakDetector:
    """
    Context manager to detect memory leaks on the given allocator.

    When we enter the context manager, we record the current sequence
    number. Upon exit, we check that all the new allocations have been freed.
    """

    def __init__(self, allocator):
        self.allocator = allocator
        self.start_seq = None

    def __enter__(self):
        if self.start_seq is not None:
            raise ValueError('LeakDetector already active')
        self.start_seq = self.allocator.seq

    def __exit__(self, etype, evalue, tb):
        leaks = []
        for addr, seq in self.allocator.alive_memory.items():
            if seq > self.start_seq:
                leaks.append((addr, seq))
        self.start_seq = None
        if leaks:
            leaks.sort(key=lambda t: t[1])  # sort by seq
            raise MemoryLeakError(leaks)


# global singleton
_ALLOCATOR = TracingAllocator()


def new_leak_detector():
    """
    Return a new instance of LeakDetector associated to the global tracing
    allocator
    """
    return LeakDetector(_ALLOCATOR)


@ffi.def_extern()
def rbclib_tracing_allocate_varlen_buffer(element_count, element_size):
    addr = lib.rbclib_allocate_varlen_buffer(element_count, element_size)
    _ALLOCATOR.record_allocate(addr)
    return addr


@ffi.def_extern()
def rbclib_tracing_free_buffer(addr):
    _ALLOCATOR.record_free(addr)
    lib.rbclib_free_buffer(addr)



"""Implement Omnisci Column type support

Omnisci Column type is the type of input/output column arguments in
UDTFs.
"""

__all__ = ['OutputColumn', 'Column', 'OmnisciOutputColumnType', 'OmnisciColumnType',
           'OmnisciCursorType']

from llvmlite import ir
from rbc import typesystem, irutils
from .omnisci_buffer import Buffer, OmnisciBufferType, BufferType
from .omnisci_column_list import OmnisciColumnListType
from rbc.targetinfo import TargetInfo
from numba.core import extending, types


int32_t = ir.IntType(32)


class OmnisciColumnType(OmnisciBufferType):
    """Omnisci Column type for RBC typesystem.
    """
    @property
    def pass_by_value(self):
        omnisci_version = TargetInfo().software[1][:3]
        return omnisci_version <= (5, 7, 0)

    def match(self, other):
        if type(self) is type(other):
            return self[0] == other[0]


class OmnisciOutputColumnType(OmnisciColumnType):
    """Omnisci OutputColumn type for RBC typesystem.

    OutputColumn is the same as Column but introduced to distinguish
    the input and output arguments of UDTFs.
    """


class Column(Buffer):
    pass


class OutputColumn(Column):
    pass


@extending.intrinsic
def omnisci_column_set_null_(typingctx, col_var, row_idx):
    # Float values are serialized as integers by OmniSciDB
    # For reference, here is the conversion table for float and double
    #   FLOAT:  1.1754944e-38            -> 8388608
    #   DOUBLE: 2.2250738585072014e-308  -> 4503599627370496
    #                    ^                          ^
    #                 fp value                  serialized
    T = col_var.eltype
    sig = types.void(col_var, row_idx)

    target_info = TargetInfo()
    null_value = target_info.null_values[str(T)]

    def codegen(context, builder, signature, args):
        data, index = args
        ptr = irutils.get_member_value(builder, data, 0)

        ty = ptr.type.pointee
        nv = ir.Constant(ir.IntType(T.bitwidth), null_value)
        if isinstance(T, types.Float):
            nv = builder.bitcast(nv, ty)
        builder.store(nv, builder.gep(ptr, [index]))

    return sig, codegen


@extending.overload_method(BufferType, 'set_null')
def omnisci_column_set_null(col_var, index):
    def impl(col_var, index):
        return omnisci_column_set_null_(col_var, index)
    return impl


@extending.intrinsic
def omnisci_column_is_null_(typingctx, col_var, row_idx):
    T = col_var.eltype
    sig = types.boolean(col_var, row_idx)

    target_info = TargetInfo()
    null_value = target_info.null_values[str(T)]
    nv = ir.Constant(ir.IntType(T.bitwidth), null_value)

    def codegen(context, builder, signature, args):
        data, index = args
        ptr = irutils.get_member_value(builder, data, 0)
        res = builder.load(builder.gep(ptr, [index]))

        if isinstance(T, types.Float):
            res = builder.bitcast(res, nv.type)

        return builder.icmp_signed('==', res, nv)

    return sig, codegen


@extending.overload_method(BufferType, 'is_null')
def omnisci_column_is_null(col_var, row_idx):
    def impl(col_var, row_idx):
        return omnisci_column_is_null_(col_var, row_idx)
    return impl


class OmnisciCursorType(typesystem.Type):

    @classmethod
    def preprocess_args(cls, args):
        assert len(args) == 1
        params = []
        for p in args[0]:
            if not isinstance(p, (OmnisciColumnType, OmnisciColumnListType)):
                # map Cursor<T ...> to Cursor<Column<T> ...>
                c = p.copy()
                p = OmnisciColumnType((c,), **c._params)
                c._params.clear()
            params.append(p)
        return (tuple(params),)

    @property
    def as_consumed_args(self):
        return self[0]



import pytest
import numpy as np
from rbc.tests import omnisci_fixture
from rbc.stdlib import array_api
from collections import OrderedDict


@pytest.fixture(scope='module')
def omnisci():

    for o in omnisci_fixture(globals(), load_test_data=False):
        define(o)
        yield o


def define(omnisci):
    @omnisci('float64(int32)')
    def get_constant(typ):
        if typ == 0:
            return array_api.e
        if typ == 1:
            return array_api.inf
        if typ == 2:
            return array_api.nan
        return array_api.pi


constants_map = OrderedDict(e=np.e, inf=np.inf, nan=np.nan, pi=np.pi)


@pytest.mark.parametrize('C', constants_map)
def test_constants(omnisci, C):
    idx = list(constants_map.keys()).index(C)
    _, result = omnisci.sql_execute(f'select get_constant({idx});')

    expected = constants_map[C]
    if np.isnan(expected):
        assert np.isnan(list(result)[0])
    else:
        assert list(result) == [(expected,)]



"""
Array API specification for creation functions.

https://data-apis.org/array-api/latest/API_specification/creation_functions.html
"""

from rbc import typesystem
from rbc.omnisci_backend.omnisci_array import Array, ArrayPointer
from rbc.stdlib import Expose
from numba import njit
from numba.core import extending, types

__all__ = [
    'full', 'full_like', 'empty_like', 'empty', 'zeros', 'zeros_like',
    'ones', 'ones_like', 'array', 'cumsum'
]


expose = Expose(globals(), 'creation_functions')


@expose.not_implemented('arange')
def arange(start, stop=None, step=1, dtype=None, device=None):
    """
    Return evenly spaced values within a given interval.
    """
    pass


@expose.not_implemented('asarray')
def asarray(obj, dtype=None, device=None, copy=None):
    """
    Convert the input to an array.
    """
    pass


@expose.not_implemented('eye')
def eye(n_rows, n_cols=None, k=0, dtype=None, device=None):
    """
    Return a 2-D array with ones on the diagonal and zeros elsewhere.
    """
    pass


@expose.not_implemented('from_dlpack')
def from_dlpack(x):
    """
    """
    pass


@expose.not_implemented('linspace')
def linspace(start, stop, num, dtype=None, device=None, endpoint=True):
    """
    Return evenly spaced numbers over a specified interval.
    """
    pass


@expose.not_implemented('meshgrid')
def meshgrid(*arrays, indexing='xy'):
    """
    Return coordinate matrices from coordinate vectors.
    """
    pass


@expose.not_implemented('tril')
def tril(x, k=0):
    """
    Lower triangle of an array.
    """
    pass


@expose.not_implemented('triu')
def triu(x, k=0):
    """
    Upper triangle of an array.
    """
    pass


@expose.implements('full')
def _omnisci_np_full(shape, fill_value, dtype=None):
    """
    Return a new array of given shape and type, filled with fill_value.
    """

    # XXX: dtype should be infered from fill_value
    if dtype is None:
        nb_dtype = types.double
    else:
        nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

    def impl(shape, fill_value, dtype=None):
        a = Array(shape, nb_dtype)
        a.fill(nb_dtype(fill_value))
        return a
    return impl


@expose.implements('full_like')
def _omnisci_np_full_like(a, fill_value, dtype=None):
    """
    Return a full array with the same shape and type as a given array.
    """
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

        def impl(a, fill_value, dtype=None):
            sz = len(a)
            other = Array(sz, nb_dtype)
            other.fill(nb_dtype(fill_value))
            return other
        return impl


@expose.implements('empty_like')
def _omnisci_np_empty_like(a, dtype=None):
    """
    Return a new array with the same shape and type as a given array.
    """
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

        def impl(a, dtype=None):
            return empty(len(a), nb_dtype)  # noqa: F821
        return impl


@expose.implements('empty')
def _omnisci_np_empty(shape, dtype=None):
    """
    Return a new array of given shape and type, without initializing entries.
    """
    if dtype is None:
        nb_dtype = types.double
    else:
        nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

    def impl(shape, dtype=None):
        arr = Array(shape, nb_dtype)
        for i in range(shape):
            arr.set_null(i)
        return arr
    return impl


@expose.implements('zeros')
def _omnisci_np_zeros(shape, dtype=None):
    """
    Return a new array of given shape and type, filled with zeros.
    """

    if dtype is None:
        nb_dtype = types.double
    else:
        nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

    fill_value = False if isinstance(nb_dtype, types.Boolean) else 0

    def impl(shape, dtype=None):
        return full(shape, fill_value, nb_dtype)  # noqa: F821
    return impl


@expose.implements('zeros_like')
def _omnisci_np_zeros_like(a, dtype=None):
    """
    Return an array of zeros with the same shape and type as a given array.
    """
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

        fill_value = False if isinstance(nb_dtype, types.Boolean) else 0

        def impl(a, dtype=None):
            return full_like(a, fill_value, nb_dtype)  # noqa: F821
        return impl


@expose.implements('ones')
def _omnisci_np_ones(shape, dtype=None):
    """
    Return a new array of given shape and type, filled with ones.
    """

    if dtype is None:
        nb_dtype = types.double
    else:
        nb_dtype = typesystem.Type.fromobject(dtype).tonumba()

    fill_value = True if isinstance(nb_dtype, types.Boolean) else 1

    def impl(shape, dtype=None):
        return full(shape, fill_value, nb_dtype)  # noqa: F821
    return impl


@expose.implements('ones_like')
def _omnisci_np_ones_like(a, dtype=None):
    """
    Return an array of ones with the same shape and type as a given array.
    """
    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = dtype

        fill_value = True if isinstance(nb_dtype, types.Boolean) else 1

        def impl(a, dtype=None):
            return full_like(a, fill_value, nb_dtype)  # noqa: F821
        return impl


@expose.implements('array')
def _omnisci_np_array(a, dtype=None):
    """
    Create an array.
    """

    @njit
    def _omnisci_array_non_empty_copy(a, nb_dtype):
        """Implement this here rather than inside "impl".
        LLVM DCE pass removes everything if we implement stuff inside "impl"
        """
        other = Array(len(a), nb_dtype)
        for i in range(len(a)):
            other[i] = a[i]
        return other

    if isinstance(a, ArrayPointer):
        if dtype is None:
            nb_dtype = a.eltype
        else:
            nb_dtype = dtype

        def impl(a, dtype=None):
            if a.is_null():
                return empty_like(a)  # noqa: F821
            else:
                return _omnisci_array_non_empty_copy(a, nb_dtype)
        return impl


@extending.overload_method(ArrayPointer, 'fill')
def _omnisci_array_fill(x, v):
    """
    Fill the array with a scalar value.
    """
    if isinstance(x, ArrayPointer):
        def impl(x, v):
            for i in range(len(x)):
                x[i] = v
        return impl


@expose.implements('cumsum')
def _omnisci_np_cumsum(a):
    """
    Return the cumulative sum of the elements along a given axis.
    """
    if isinstance(a, ArrayPointer):
        eltype = a.eltype

        def impl(a):
            sz = len(a)
            out = Array(sz, eltype)
            out[0] = a[0]
            for i in range(sz):
                out[i] = out[i-1] + a[i]
            return out
        return impl



"""
https://data-apis.org/array-api/latest/API_specification/constants.html
"""
import numpy as np


__all__ = [
    'e', 'inf', 'nan', 'pi'
]


# it doesn't seem to be possible to document constants with autosummary
# https://github.com/sphinx-doc/sphinx/issues/6794
e = np.e
inf = np.inf
nan = np.nan
pi = np.pi



import os
import ctypes
import ctypes.util
from enum import IntEnum
from .utils import runcommand

#
# Import CUDA RT library
#

if os.name == 'nt':
    lib = ctypes.util.find_library('libcudart')
    if lib is None:
        lib = ctypes.util.find_library('libcudart.dll')
else:
    lib = ctypes.util.find_library('cudart')

if lib is None:
    raise ImportError('Failed to find CUDA RT library.'
                      ' Make sure that the library is installed')


def get_cuda_libdir():
    for line in runcommand('ldconfig', '-p').splitlines():
        if 'cudart' in line:
            libpath = os.path.realpath(line.split('=>')[-1].strip())
            return os.path.dirname(libpath)
    return os.path.dirname(lib)


if not os.path.isfile(lib):
    lib = os.path.join(get_cuda_libdir(), lib)

libdir = os.path.dirname(lib)
includedir = os.path.join(os.path.dirname(libdir), 'include')

libcudart = ctypes.cdll.LoadLibrary(lib)

#
# Auxiliary functions for reading CUDA header files
#


def get_cuda_header(headername, _cache={}):
    content = _cache.get(headername)
    if content is None:
        content = open(os.path.join(includedir, headername)).read()
        _cache[headername] = content
    return content


def read_cuda_header(headername, kind, name, process_func):
    content = get_cuda_header(headername)
    flag = False
    d = {}
    key = -1
    for line in content.splitlines():
        line = line.strip()
        if not flag and line.startswith(kind) and line.endswith(name):
            flag = True
            comment = False
        elif flag:
            if line.startswith('}'):
                break
            if comment:
                j = line.find('*/')
                if j != -1:
                    comment = False
                    line = line[j+2:]
                else:
                    continue
            else:
                i = line.find('/*')
                if i != -1:
                    j = line.find('*/')
                    if j != -1:
                        line = line[:i] + line[j+2:]
                    else:
                        comment = True
                        continue

            line = line.strip()
            if line.endswith(',') or line.endswith(';'):
                line = line[:-1].rstrip()
            if not line or line == '{':
                continue
            key, value = process_func(line, key)
            if value is None:
                continue
            d[key] = value
    return d


def get_cuda_enum(headerfile, name):
    """Return value:name mapping of a enum definition.
    """
    def process_enum(line, last_value):
        if '=' in line:
            n, v = line.split('=')
            n = n.strip()
            v = eval(v.strip(), {}, {})
        else:
            n = line
            v = last_value + 1
        return v, n
    return read_cuda_header(headerfile, 'enum', name, process_enum)


def get_cuda_struct(headerfile, name):
    """Return membername:type mapping of a struct definition.
    """
    def process_struct(line, last_value):
        typ, decl = line.rsplit(' ', 1)
        typ = typ.strip()
        t = get_ctype(typ)
        i = decl.find('[')
        if i != -1:
            dim = int(decl[i+1:-1])
            t = t * dim
            n = decl[:i]
        else:
            n = decl
        return n, t
    return read_cuda_header(headerfile, 'struct', name, process_struct)


def get_ctype(tstr):
    t = getattr(ctypes, 'c_' + tstr, None)
    if t is not None:
        return t
    t = {'cudaUUID_t': cudaUUID_t,
         'unsigned int': ctypes.c_uint}.get(tstr)
    if t is not None:
        return t
    raise NotImplementedError('get ctypes version of `%s`' % (tstr))


#
# CUDA enum definitions
#


class CTypesEnum(IntEnum):

    # See
    #   https://www.chriskrycho.com/2015/ctypes-structures-and-dll-exports.html
    @classmethod
    def from_param(cls, obj):
        return int(obj)


cudaError_value_name_map = get_cuda_enum('driver_types.h', 'cudaError')
cudaError_name_value_map = dict((n, v)
                                for (v, n) in cudaError_value_name_map.items())
_cudaError_member_defs = '\n'.join(
    ['%s = %s' % _item for _item in cudaError_name_value_map.items()])


class cudaError(CTypesEnum):
    exec(_cudaError_member_defs)


#
# CUDA struct definitions
#


class cudaUUID_t(ctypes.Structure):
    _fields_ = [('bytes', ctypes.c_char * 16)]


class cudaDeviceProp(ctypes.Structure):
    _fields_ = list(get_cuda_struct('driver_types.h',
                                    'cudaDeviceProp').items())


#
# CUDA functions
#

libcudart.cudaDriverGetVersion.argtypes = [ctypes.POINTER(ctypes.c_int)]
libcudart.cudaDriverGetVersion.restype = cudaError

libcudart.cudaRuntimeGetVersion.argtypes = [ctypes.POINTER(ctypes.c_int)]
libcudart.cudaRuntimeGetVersion.restype = cudaError

libcudart.cudaGetDeviceProperties.argtypes = [ctypes.POINTER(cudaDeviceProp),
                                              ctypes.c_int]
libcudart.cudaGetDeviceProperties.restype = cudaError

libcudart.cudaGetDeviceCount.argtypes = [ctypes.POINTER(ctypes.c_int)]
libcudart.cudaGetDeviceCount.restype = cudaError

#
# Convinience functions
#


def get_cuda_versions():
    """Return CUDA driver and runtime versions.
    """
    dr = ctypes.c_int(0)
    rt = ctypes.c_int(0)
    r = libcudart.cudaDriverGetVersion(ctypes.byref(dr))
    assert r == cudaError.cudaSuccess, repr(r)
    if dr.value != 0:
        # driver is installed
        r = libcudart.cudaRuntimeGetVersion(ctypes.byref(rt))
        assert r in [cudaError.cudaSuccess,
                     cudaError.cudaErrorNoDevice], repr(r)
    return dr.value, rt.value


def get_cuda_device_properties(device):
    """Return CUDA device properties as a dictionary.

    The value types in the dictionary will be int, bytes, or list.
    """
    v = cudaDeviceProp()
    d = ctypes.c_int(0)
    r = libcudart.cudaGetDeviceProperties(ctypes.byref(v), d)
    assert r in [cudaError.cudaSuccess, cudaError.cudaErrorNoDevice], repr(r)
    d = {}
    for (n, t) in cudaDeviceProp._fields_:
        a = getattr(v, n)
        if isinstance(a, ctypes.Array):
            a = list(a)
        d[n] = a
    return d


def get_device_count():
    """Return the number of CUDA devices.
    """
    c = ctypes.c_int(0)
    r = libcudart.cudaGetDeviceCount(ctypes.byref(c))
    if r == cudaError.cudaErrorNoDevice:
        return 0
    assert r == cudaError.cudaSuccess, repr(r)  # noqa: F821
    return c.value



"""
rbc-specific errors and warnings.
"""


from rbc.utils import get_version
from numba.core.errors import TypingError


class OmnisciServerError(Exception):
    """
    Raised when OmnisciDB server raises a runtime error that RBC knows
    how to interpret.
    """
    pass


class UnsupportedError(Exception):
    """
    Raised when an attempt to use a feature that is not supported for
    a given target. The attempt may be made both in lowering as well
    as in typing phase.
    """
    def __init__(self, *args, **kwargs):
        # numba mangles exception messages. Here we insert the exception
        # name so that irtools can demangle the messages.
        Exception.__init__(self, type(self).__name__, *args, **kwargs)


class ForbiddenNameError(Exception):
    """
    Raised when the user defines a function with name
    in a blacklist. For more info, see:
    https://github.com/xnd-project/rbc/issues/32
    """
    pass


class ForbiddenIntrinsicError(Exception):
    """
    Raised when the user uses an intrinsic with name
    in a blacklist. For more info, see:
    https://github.com/xnd-project/rbc/issues/207
    """
    pass


if get_version('numba') < (0, 55):
    class NumbaTypeError(TypingError):
        pass

    class NumbaNotImplementedError(TypingError):
        pass
else:
    from numba.core.errors import NumbaTypeError, NumbaNotImplementedError  # noqa: F401



"""RemoteJIT client/server config functions
"""

__all__ = ['RemoteJIT', 'Signature', 'Caller']

import os
import inspect
import warnings
import ctypes
from contextlib import nullcontext
from collections import defaultdict
from . import irtools
from .errors import UnsupportedError
from .typesystem import Type, get_signature
from .thrift import Server, Dispatcher, dispatchermethod, Data, Client
from .utils import get_local_ip, UNSPECIFIED
from .targetinfo import TargetInfo
from .rbclib import tracing_allocator
# XXX WIP: the OmnisciCompilerPipeline is no longer omnisci-specific because
# we support Arrays even without omnisci, so it must be renamed and moved
# somewhere elsef
from .omnisci_backend import OmnisciCompilerPipeline


def isfunctionlike(obj):
    """Return True if object is function alike.
    """
    if obj is None or isinstance(obj, (Signature, list, tuple, str, Caller)):
        return False
    return True


def extract_templates(options):
    """Extract templates mapping data from options dictionary.

    If options does not contain "templates", it will be constructed
    from all unknown options that have list values. Otherwise, the
    corresponding value is returned with no further processing of
    options content.

    Parameters
    ----------
    options : dict

    Returns
    -------
    options : dict
      A copy of input without templates mapping data.
    templates : dict
      Templates mapping which is a collections of pairs of template
      name and a list of concrete types. Template name cannot
      correspond to a concrete type.

    """
    known_options = ['devices', 'local']
    new_options = {}
    templates = options.get('templates')
    if templates is not None:
        new_options.update(options)
        del new_options['templates']
    else:
        templates = {}
        for k, v in options.items():
            if (isinstance(k, str) and isinstance(v, list) and k not in known_options):
                templates[k] = v
            else:
                new_options[k] = v
    return new_options, templates


class Signature:
    """Signature decorator for Python functions.

    A Signature decorator may contain many signature objects
    representing the prototypes of functions.

    Signature decorators are re-usable and composeable. For example:

    .. highlight:: python
    .. code-block:: python

        rjit = RemoteJIT(host='localhost' port=6274)

        # remotebinaryfunc is Signature instance
        remotebinaryfunc = rjit('int32(int32, int32)',
                                'float32(float32, float32)', ...)

        # add will be Caller instance
        @remotebinaryfunc
        def add(a, b):
            return a + b

        # sub will be Caller instance
        @remotebinaryfunc
        def sub(a, b):
            return a - b

        add(1, 2) # returns 3
        sub(1.0, 2.0) # returns -1.0
    """

    def __init__(self, remotejit):
        assert isinstance(remotejit, RemoteJIT), type(remotejit)
        self.remotejit = remotejit
        self.signatures = []
        self.signature_devices = {}
        self.signature_templates = {}

    @property
    def debug(self):
        return self.remotejit.debug

    @property
    def local(self):
        sig = Signature(self.remotejit.local)
        sig.signatures.extend(self.signatures)
        assert not self.signature_devices
        assert not self.signature_templates
        return sig

    def __repr__(self):
        return f'{type(self).__name__}({str(self)})'

    def __str__(self):
        lst = []
        for t in self.signatures:
            s = str(t)
            for k, types in self.signature_templates.get(t, {}).items():
                s += f', {k}={"|".join(map(str, types))}'
            devices = self.signature_devices.get(t, [])
            if devices:
                s += f', device={"|".join(devices)}'
            lst.append(repr(s))
        return f'{"; ".join(lst)}'

    def __call__(self, obj, **options):
        """Decorate signatures or a function.

        Parameters
        ----------
        obj : {str, Signature, function, ...}
          Specify object that represents a function type.

        Keyword parameters
        ------------------
        devices : list
          Specify device names for the given set of signatures.
        templates : dict
          Specify template types mapping.

        Returns
        -------
        result : {Signature, Caller}
          If obj is a function, return Caller. Otherwise return self
          that is extended with new signatures from obj.

        Note
        ----
        The validity of the input argument is not checked here.  This
        is because the bit-size of certain C types (e.g. size_t, long,
        etc) depend on the target device which information will be
        available at the compile stage. The target dependent
        signatures can be retrieved using
        `signature.get_signatures()`.
        """
        if obj is None:
            return self
        options, templates = extract_templates(options)
        devices = options.get('devices')
        if isinstance(obj, Signature):
            self.signatures.extend(obj.signatures)
            self.signature_devices.update(obj.signature_devices)
            self.remotejit.discard_last_compile()
            if devices is not None:
                for s in obj.signatures:
                    self.signature_devices[s] = devices
            assert not templates
            for s in obj.signatures:
                t = obj.signature_templates.get(s)
                if t is not None:
                    self.signature_templates[s] = t
            return self
        if isinstance(obj, Caller):
            # return new Caller with extended signatures set
            assert obj.remotejit is self.remotejit
            final = Signature(self.remotejit)
            final(self)  # copies the signatures from self to final
            final(obj.signature)  # copies the signatures from obj to final
            assert devices is None
            assert not templates
            return Caller(obj.func, final)
        if isfunctionlike(obj):
            final = Signature(self.remotejit)
            final(self)  # copies the signatures from self to final
            assert devices is None
            assert not templates
            return Caller(obj, final)
        self.signatures.append(obj)
        self.remotejit.discard_last_compile()
        if devices is not None:
            self.signature_devices[obj] = devices
        if templates:
            self.signature_templates[obj] = templates
        return self

    def best_match(self, func, atypes: tuple) -> Type:
        """Return function type from signatures that matches best with given
        argument types.

        If no match is found, raise TypeError.

        Parameters
        ----------
        atypes : Type-tuple
          Specify a tuple of argument types.

        Returns
        -------
        ftype : Type
          Function type that arguments match best with given argument
          types.
        """
        ftype = None
        match_penalty = None
        available_types = self.normalized(func).signatures

        for typ in available_types:
            sig = self.remotejit.caller_signature(typ)
            penalty = sig.match(atypes)
            if penalty is not None:
                if ftype is None or penalty < match_penalty:
                    ftype = typ
                    match_penalty = penalty
        return ftype, match_penalty

    def add(self, sig):
        if sig not in self.signatures:
            self.signatures.append(sig)

    def normalized(self, func=None):
        """Return a copy of Signature object where all signatures are
        normalized to Type instances using the current target device
        information.

        Parameters
        ----------
        func : {None, callable}
          Python function that annotations are attached to signature.

        Returns
        -------
        signature : Signature
        """
        signature = Signature(self.remotejit)
        fsig = Type.fromcallable(func) if func is not None else None
        nargs = fsig.arity if func is not None else None
        target_info = TargetInfo()
        for sig in self.signatures:
            devices = self.signature_devices.get(sig)
            if not target_info.check_enabled(devices):
                if self.debug:
                    print(f'{type(self).__name__}.normalized: skipping {sig} as'
                          f' not supported by devices: {devices}')
                continue
            templates = self.signature_templates.get(sig, {})
            sig = Type.fromobject(sig)
            if not sig.is_complete:
                warnings.warn(f'Incomplete signature {sig} will be ignored')
                continue
            if not sig.is_function:
                raise ValueError(
                    'expected signature representing function type,'
                    f' got `{sig}`')
            if nargs is None:
                nargs = sig.arity
            elif sig.arity != nargs:
                raise ValueError(f'signature `{sig}` must have arity {nargs}'
                                 f' but got {len(sig[1])}')
            if fsig is not None:
                sig.inherit_annotations(fsig)

            if not sig.is_concrete:
                for csig in sig.apply_templates(templates):
                    assert isinstance(csig, Type), (sig, csig, type(csig))
                    csig = self.remotejit.normalize_function_type(csig)
                    signature.add(csig)
            else:
                sig = self.remotejit.normalize_function_type(sig)
                signature.add(sig)
        if fsig is not None and fsig.is_complete:
            fsig = self.remotejit.normalize_function_type(fsig)
            signature.add(fsig)
        return signature


class Caller:
    """Remote JIT caller, holds the decorated function that can be
    executed remotely.
    """

    def __init__(self, func, signature: Signature):
        """Construct remote JIT caller instance.

        Parameters
        ----------
        func : callable
          Specify a Python function that is used as a template to
          remotely JIT compiled functions.
        signature : Signature
          Specify a collection of signatures.
        local : bool
          When True, local process will be interpreted as
          remote. Useful for debugging.
        """
        self.remotejit = signature.remotejit
        self.signature = signature
        func = self.remotejit.preprocess_callable(func)
        self.func = func
        self.nargs = len(get_signature(func).parameters)

        # Attributes used in RBC user-interface
        self._is_compiled = set()  # items are (fname, ftype)
        self._client = None

        self.remotejit.add_caller(self)

    @property
    def local(self):
        """Return Caller instance that executes function calls on the local
        host. Useful for debugging.
        """
        if not self.remotejit.supports_local_caller:
            msg = (
                "Cannot create a local `Caller` when using "
                f"{type(self.remotejit).__name__}."
            )
            raise UnsupportedError(msg)

        return Caller(self.func, self.signature.local)

    def __repr__(self):
        return f"{type(self).__name__}({self.func}, {self.signature!r})"

    def __str__(self):
        return f"{self.func.__name__}[{self.signature}]"

    def describe(self):
        """Return LLVM IRs of all target devices.
        """
        lst = ['']
        fid = 0
        for device, target_info in self.remotejit.targets.items():
            with Type.alias(**self.remotejit.typesystem_aliases):
                with target_info:
                    lst.append(f'{device:-^80}')
                    signatures = self.get_signatures()
                    signatures_map = {}
                    for sig in signatures:
                        fid += 1
                        signatures_map[fid] = sig
                    llvm_module, succesful_fids = irtools.compile_to_LLVM(
                        [(self.func, signatures_map)],
                        target_info,
                        pipeline_class=OmnisciCompilerPipeline,
                        debug=self.remotejit.debug)
                    lst.append(str(llvm_module))
        lst.append(f'{"":-^80}')
        return '\n'.join(lst)

    def get_signatures(self):
        """Return a list of normalized signatures for given target device.
        """
        return self.signature.normalized(self.func).signatures

    # RBC user-interface

    def __call__(self, *arguments, device=UNSPECIFIED, hold=UNSPECIFIED):
        """Return the result of a remote JIT compiled function call.
        """
        caller = self.remotejit.get_caller(self.func.__name__)
        return caller(*arguments, device=device, hold=hold)


class RemoteDispatcher:
    """A collection of Caller instances holding functions with a common name.
    """
    def __init__(self, name, callers):
        self.name = name
        assert callers  # at least one caller must be specified
        self.remotejit = callers[0].remotejit
        self.callers = callers

    def __repr__(self):
        lst = [str(caller.signature) for caller in self.callers]
        return f'{type(self).__name__}({self.name!r}, [{", ".join(lst)}])'

    def __str__(self):
        lst = [str(caller.signature) for caller in self.callers]
        return f'{self.name}[{", ".join(lst)}]'

    def __call__(self, *arguments, device=UNSPECIFIED, hold=UNSPECIFIED):
        """Perform remote call with given arguments.

        If `hold` is True, return an object that encapsulates the
        remote call to postpone the remote execution.
        """
        if hold is UNSPECIFIED:
            hold = self.remotejit.default_remote_call_hold

        penalty_device_caller_ftype = []
        atypes = None
        for device_, target_info in self.remotejit.targets.items():
            if device is not UNSPECIFIED and device != device_:
                continue
            with target_info:
                atypes = self.remotejit.get_types(*arguments)
                for caller_id, caller in enumerate(self.callers):
                    with Type.alias(**self.remotejit.typesystem_aliases):
                        ftype, penalty = caller.signature.best_match(caller.func, atypes)
                        if ftype is None:
                            continue
                        penalty_device_caller_ftype.append((penalty, device_, caller_id, ftype))
        if atypes is None:
            raise ValueError(f'no target info found for given device {device}')

        penalty_device_caller_ftype.sort()

        if not penalty_device_caller_ftype:
            available_types_devices = defaultdict(set)
            for device_, target_info in self.remotejit.targets.items():
                if device is not UNSPECIFIED and device != device_:
                    continue
                with target_info:
                    for caller in self.callers:
                        with Type.alias(**self.remotejit.typesystem_aliases):
                            for t in caller.signature.normalized(caller.func).signatures:
                                available_types_devices[t].add(device_)
            lines = self.remotejit._format_available_function_types(available_types_devices)
            available = '\n    ' + '\n    '.join(lines)
            satypes = ', '.join(map(str, atypes))
            raise TypeError(
                f'found no matching function signature to given argument types:'
                f'\n    ({satypes}) -> ...\n  available function signatures:{available}')

        _, device, caller_id, ftype = penalty_device_caller_ftype[0]
        target_info = self.remotejit.targets[device]
        caller = self.callers[caller_id]
        r = self.remotejit.remote_call_capsule_cls(caller, target_info, ftype, arguments)

        return r if hold else r.execute()


class RemoteCallCapsule:
    """Encapsulates remote call execution.
    """

    use_execute_cache = False

    def __init__(self, caller, target_info, ftype, arguments):
        self.caller = caller
        self.target_info = target_info
        self.ftype = ftype
        self.arguments = arguments
        self._execute_cache = UNSPECIFIED

    @property
    def __typesystem_type__(self):
        """The typesystem Type instance of the return value of the remote
        call.
        """
        return self.caller.remotejit.caller_signature(self.ftype)[0]

    def __repr__(self):
        return (f'{type(self).__name__}({self.caller!r}, {self.target_info},'
                f' {self.ftype}, {self.arguments})')

    def __str__(self):
        return f'{self.execute(hold=True)}'

    def execute(self, hold=False):
        """Trigger the remote call execution.

        When `hold` is True, return an object that represents the
        remote call.
        """
        if not hold:
            if self.use_execute_cache and self._execute_cache is not UNSPECIFIED:
                return self._execute_cache

            key = self.caller.func.__name__, self.ftype
            if key not in self.caller._is_compiled:
                self.caller.remotejit.remote_compile(
                    self.caller.func, self.ftype, self.target_info)
                self.caller._is_compiled.add(key)
        result = self.caller.remotejit.remote_call(self.caller.func, self.ftype,
                                                   self.arguments, hold=hold)
        if not hold and self.use_execute_cache:
            self._execute_cache = result
        return result


class RemoteJIT:
    """RemoteJIT is a decorator generator for user functions to be
    remotely JIT compiled.

    To use, define

    .. highlight:: python
    .. code-block:: python

        rjit = RemoteJIT(host='localhost', port=6274)

        @rjit
        def foo(a: int, b: int) -> int:
            return a + b

        @rjit('double(double, double)',
              'int64(int64, int64)')
        def bar(a, b):
            return a + b

        # Finally, call
        c = foo(1, 2) # c = 3
        b = bar(7.0, 1.0) # b = 8.0

    The sum will be evaluated in the remote host.
    """

    multiplexed = True

    thrift_content = None

    typesystem_aliases = dict()

    remote_call_capsule_cls = RemoteCallCapsule

    # Some callers cannot be called locally
    supports_local_caller = True

    # Should calling RemoteDispatcher hold the execution:
    default_remote_call_hold = False

    def __init__(self, host='localhost', port=11532,
                 local=False, debug=False, use_tracing_allocator=False):
        """Construct remote JIT function decorator.

        The decorator is re-usable for different functions.

        Parameters
        ----------
        host : str
          Specify the host name of IP of JIT server
        port : {int, str}
          Specify the service port of the JIT server
        local : bool
          When True, use local client. Useful for debugging.
        debug : bool
          When True, output debug messages.
        use_tracing_allocator : bool
          When True, enable the automatic detection of memory leaks.
        """
        if host == 'localhost':
            host = get_local_ip()

        if use_tracing_allocator and not local:
            raise ValueError('use_tracing_allocator=True can be used only with local=True')

        self.debug = debug
        self.use_tracing_allocator = use_tracing_allocator
        self.host = host
        self.port = int(port)
        self.server_process = None

        # A collection of Caller instances. Each represents a function
        # that have many argument type dependent implementations.
        self._callers = []
        self._last_compile = None
        self._targets = None

        if local:
            self._client = LocalClient(debug=debug,
                                       use_tracing_allocator=use_tracing_allocator)
        else:
            self._client = None

    @property
    def local(self):
        localjit = type(self)(local=True, debug=self.debug)
        localjit._callers.extend(self._callers)
        return localjit

    def add_caller(self, caller):
        name = caller.func.__name__
        for c in self._callers:
            if c.name == name:
                c.callers.append(caller)
                break
        else:
            self._callers.append(RemoteDispatcher(name, [caller]))
        self.discard_last_compile()

    def get_callers(self):
        callers = []
        for c in self._callers:
            callers.extend(c.callers)
        return callers

    def get_caller(self, name):
        for c in self._callers:
            if c.name == name:
                return c
        return

    def reset(self):
        """Drop all callers definitions and compilation results.
        """
        self._callers.clear()
        self.discard_last_compile()

    @property
    def have_last_compile(self):
        """Check if compile data exists.

        See `set_last_compile` method for more information.
        """
        return self._last_compile is not None

    def discard_last_compile(self):
        """Discard compile data.

        See `set_last_compile` method for more information.
        """
        self._last_compile = None

    def set_last_compile(self, compile_data):
        """Save compile data.

        The caller is responsible for discarding previous compiler
        data by calling `discard_last_compile` method.

        Parameters
        ----------
        compile_data : object
          Compile data can be any Python object. When None, it is
          interpreted as no compile data is available.


        Notes
        -----

        The have/discard/set_last_compile methods provide a way to
        avoid unnecessary compilations when the remote server supports
        registration of compiled functions. The corresponding
        `register` method is expected to use the following pattern:


        .. code-block:: python

           def register(self):
               if self.have_last_compile:
                   return
               <compile defined functions>
               self.set_last_compile(<compilation results>)


        The `discard_last_compile()` method is called when the compile
        data becomes obsolete or needs to be discarded. For instance,
        the compile data will be discarded when calling the following
        methods: `reset`, `add_caller`. Note that the `add_caller`
        call is triggered when applying the remotejit decorator to a
        Python function to be compiled.

        """
        assert self._last_compile is None
        self._last_compile = compile_data

    def get_pending_names(self):
        """Return the names of functions that have not been registered to the
        remote server.
        """
        names = set()
        if not self.have_last_compile:
            for caller in reversed(self.get_callers()):
                names.add(caller.func.__name__)
        return names

    def retrieve_targets(self):
        """Retrieve target device information from remote client.

        Redefine this method if remote client is not native.

        Returns
        -------
        targets : dict
          Map of target device names and informations.
        """
        # TODO: rename thrift API targets to get_device_parameters?
        response = self.client(remotejit=dict(targets=()))
        targets = {}
        for device, data in response['remotejit']['targets'].items():
            targets[device] = TargetInfo.fromjson(data)
        return targets

    @property
    def targets(self):
        """Return device-target_info mapping of the remote server.
        """
        if self._targets is None:
            self._targets = self.retrieve_targets()
        return self._targets

    def __call__(self, *signatures, devices=None, local=False, **templates):
        """Define a remote JIT function signatures and template.

        Parameters
        ----------
        signatures : str or object
          Specify signatures of a remote JIT function, or a Python
          function as a template from which the remote JIT function
          will be compiled.

        Keyword parameters
        ------------------
        local : bool
        devices : list
          Specify device names for the given set of signatures. Possible
          values are 'cpu', 'gpu'.
        templates : dict(str, list(str))
          Specify template types mapping.

        Returns
        -------
        sig: {Signature, Caller}
          Signature decorator or Caller

        Notes
        -----
        The signatures can be strings in the following form:

          "<return type>(<argument type 1>, <argument type 2>, ...)"

        or any other object that can be converted to function type,
        see `Type.fromobject` for more information.
        """
        if local:
            s = Signature(self.local)
        else:
            s = Signature(self)
        options = dict(
            local=local,
            devices=devices,
            templates=templates.get('templates') or templates
        )
        if devices is not None and not {'cpu', 'gpu'}.issuperset(devices):
            raise ValueError("'devices' can only be a list with possible "
                             f"values 'cpu', 'gpu' but got {devices}")

        _, templates = extract_templates(options)
        for sig in signatures:
            s = s(sig, devices=devices, templates=templates)
        return s

    def start_server(self, background=False):
        """Start remotejit server from client.
        """
        thrift_file = os.path.join(os.path.dirname(__file__),
                                   'remotejit.thrift')
        print('staring rpc.thrift server: %s' % (thrift_file), end='',
              flush=True)
        if self.debug:
            print(flush=True)
            dispatcher = DebugDispatcherRJIT
        else:
            dispatcher = DispatcherRJIT
        if background:
            ps = Server.run_bg(dispatcher, thrift_file,
                               dict(host=self.host, port=self.port,
                                    debug=self.debug))
            self.server_process = ps
        else:
            Server.run(dispatcher, thrift_file,
                       dict(host=self.host, port=self.port,
                            debug=self.debug))
            print('... rpc.thrift server stopped', flush=True)

    def stop_server(self):
        """Stop remotejit server from client.
        """
        if self.server_process is not None and self.server_process.is_alive():
            print('... stopping rpc.thrift server')
            self.server_process.terminate()
            self.server_process = None

    @property
    def client(self):
        """Return remote host connection as Client instance.
        """
        if self._client is None:
            self._client = Client(
                host=self.host,
                port=self.port,
                multiplexed=self.multiplexed,
                thrift_content=self.thrift_content,
                socket_timeout=60000)
        return self._client

    def remote_compile(self, func, ftype: Type, target_info: TargetInfo):
        """Remote compile function and signatures to machine code.

        The input function `func` is compiled to LLVM IR module, the
        LLVM IR module is sent to remote host where the remote host is
        expected to complete the compilation process.

        Return the corresponding LLVM IR module instance which may be
        useful for debugging.
        """
        if self.debug:
            print(f'remote_compile({func}, {ftype})')
        with target_info:
            llvm_module, succesful_fids = irtools.compile_to_LLVM(
                [(func, {0: ftype})],
                target_info,
                pipeline_class=OmnisciCompilerPipeline,
                debug=self.debug)
        ir = str(llvm_module)
        mangled_signatures = ';'.join([s.mangle() for s in [ftype]])
        response = self.client(remotejit=dict(
            compile=(func.__name__, mangled_signatures, ir)))
        assert response['remotejit']['compile'], response
        return llvm_module

    def remote_call(self, func, ftype: Type, arguments: tuple, hold=False):
        """Call function remotely on given arguments.

        The input function `func` is called remotely by sending the
        arguments data to remote host where the previously compiled
        function (see `remote_compile` method) is applied to the
        arguments, and the result is returned to local process.

        If `hold` is True then return an object that specifies remote
        call but does not execute it. The type of return object is
        custom to particular RemoteJIT specialization.
        """
        if self.debug:
            print(f'remote_call({func}, {ftype}, {arguments})')
        fullname = func.__name__ + ftype.mangle()
        call = dict(call=(fullname, arguments))
        if hold:
            return call
        response = self.client(remotejit=call)
        return response['remotejit']['call']

    def python(self, statement):
        """Execute Python statement remotely.
        """
        response = self.client(remotejit=dict(python=(statement,)))
        return response['remotejit']['python']

    def normalize_function_type(self, ftype: Type):
        """Apply RemoteJIT specific hooks to normalized function Type.

        Parameters
        ----------
        ftype: Type
          typesystem type of a function

        Returns
        -------
        ftype: Type
          typesystem type of a function with normalization hools applied
        """
        assert ftype.is_function, ftype
        return ftype

    def preprocess_callable(self, func):
        """Preprocess func to be used as a remotejit function definition.

        Parameters
        ----------
        func : callable

        Returns
        -------
        func : callable
          Preprocessed func.
        """
        return func

    def caller_signature(self, signature: Type):
        """Return signature of a caller.

        Parameters
        ----------
        signature: Type
          Signature of function implementation

        Returns
        -------
        signature: Type
          Signature of function caller
        """
        return signature

    def get_types(self, *values):
        """Convert values to the corresponding typesystem types.
        """
        return tuple(map(Type.fromvalue, values))

    def format_type(self, typ: Type):
        """Convert typesystem type to formatted string.
        """
        return str(typ)

    def _format_available_function_types(self, available_types_devices):
        all_devices = set()
        list(map(all_devices.update, available_types_devices.values()))
        lines = []
        for typ, devices in available_types_devices.items():
            sig = self.caller_signature(typ)
            d = '  ' + '|'.join(devices) + ' only' if len(devices) != len(all_devices) else ''
            s = self.format_type(sig)
            t = self.format_type(typ)
            if sig == typ:
                lines.append(f'{s}{d}')
            else:
                lines.append(f'{s}{d}\n    - {t}')
        return lines


class DispatcherRJIT(Dispatcher):
    """Implements remotejit service methods.
    """

    def __init__(self, server, debug=False, use_tracing_allocator=False):
        super().__init__(server, debug=debug)
        self.use_tracing_allocator = use_tracing_allocator
        self.compiled_functions = dict()
        self.engines = dict()
        self.python_globals = dict()
        self.python_locals = dict()

    @dispatchermethod
    def targets(self) -> dict:
        """Retrieve target device information.

        Returns
        -------
        info : dict
          Map of target devices and their properties.
        """
        if self.use_tracing_allocator:
            target_info = TargetInfo.host(name='host_cpu_tracing_allocator',
                                          use_tracing_allocator=True)
        else:
            target_info = TargetInfo.host()
        target_info.set('has_numba', True)
        target_info.set('has_cpython', True)
        return dict(cpu=target_info.tojson())

    @dispatchermethod
    def compile(self, name: str, signatures: str, ir: str) -> int:
        """JIT compile function.

        Parameters
        ----------
        name : str
          Specify the function name.
        signatures : str
          Specify semi-colon separated list of mangled signatures.
        ir : str
          Specify LLVM IR representation of the function.
        """
        engine = irtools.compile_IR(ir)
        for msig in signatures.split(';'):
            sig = Type.demangle(msig)
            ctypes_sig = sig.toctypes()
            assert sig.is_function
            if sig[0].is_aggregate:
                raise RuntimeError(
                    f'Functions with aggregate return type values are not supported,'
                    f' got function `{name}` with `{sig}` signature')
            fullname = name + msig
            addr = engine.get_function_address(fullname)
            if self.debug:
                print(f'compile({name}, {sig}) -> {hex(addr)}')
            # storing engine as the owner of function addresses
            if addr:
                self.compiled_functions[fullname] = engine, ctypes_sig(addr), sig, ctypes_sig
            else:
                warnings.warn('No compilation result for {name}|{sig=}')
        return True

    @dispatchermethod
    def call(self, fullname: str, arguments: tuple) -> Data:
        """Call JIT compiled function

        Parameters
        ----------
        fullname : str
          Specify the full name of the function that is in form
          "<name><mangled signature>"
        arguments : tuple
          Specify the arguments to the function.
        """
        # if we are using a tracing allocator, automatically detect memory leaks
        # at each call.
        if self.use_tracing_allocator:
            leak_detector = tracing_allocator.new_leak_detector()
        else:
            leak_detector = nullcontext()
        with leak_detector:
            return self._do_call(fullname, arguments)

    def _do_call(self, fullname, arguments):
        if self.debug:
            print(f'call({fullname}, {arguments})')
        ef = self.compiled_functions.get(fullname)
        if ef is None:
            raise RuntimeError(
                f'no such compiled function `{fullname}`. Available functions:\n'
                f'  {"; ".join(list(self.compiled_functions))}\n.')
        sig = ef[2]
        ctypes_sig = ef[3]
        if len(arguments) == 0:
            assert sig.arity == 1 and sig[1][0].is_void, sig
        else:
            assert len(arguments) == sig.arity, (len(arguments), sig.arity)
        ctypes_arguments = []
        for typ, ctypes_typ, value in zip(sig[1], ctypes_sig._argtypes_, arguments):
            if typ.is_custom:
                typ = typ.get_struct_type()
            if typ.is_struct:
                if isinstance(value, tuple):
                    member_values = [t.toctypes()(value[i]) for i, t in enumerate(typ)]
                else:
                    member_values = [t.toctypes()(getattr(value, t.name)) for t in typ]
                ctypes_arguments.extend(member_values)
            elif typ.is_pointer:
                if isinstance(value, ctypes.c_void_p):
                    value = ctypes.cast(value, ctypes_typ)
                else:
                    value = ctypes.cast(value, ctypes_typ)
                ctypes_arguments.append(value)
            else:
                ctypes_arguments.append(value)
        r = ef[1](*ctypes_arguments)
        if sig[0].is_pointer and sig[0][0].is_void and isinstance(r, int):
            r = ctypes.c_void_p(r)
        if self.debug:
            print(f'-> {r}')
        if hasattr(r, 'topython'):
            return r.topython()
        return r

    @dispatchermethod
    def python(self, statement: str) -> int:
        """Execute Python statement.
        """
        if self.debug:
            print(f'python({statement!r})')

        exec(statement, self.python_globals, self.python_locals)
        return True


class DebugDispatcherRJIT(DispatcherRJIT):
    """
    Enables debug messages.
    """
    debug = True


class LocalClient:
    """Pretender of thrift.Client.

    All calls will be made in a local process. Useful for debbuging.
    """

    def __init__(self, debug=False, use_tracing_allocator=False):
        self.dispatcher = DispatcherRJIT(None, debug=debug,
                                         use_tracing_allocator=use_tracing_allocator)

    def __call__(self, **services):
        results = {}
        for service_name, query_dict in services.items():
            results[service_name] = {}
            for mthname, args in query_dict.items():
                mth = getattr(self.dispatcher, mthname)
                mth = inspect.unwrap(mth)
                results[service_name][mthname] = mth(self.dispatcher, *args)
        return results



import pytest
from rbc.tests import omnisci_fixture


@pytest.fixture(scope='module')
def omnisci():

    for o in omnisci_fixture(globals(), minimal_version=(5, 6)):
        define(o)
        yield o


def define(omnisci):

    @omnisci('int32(Column<T>, RowMultiplier, OutputColumn<T>)',
             T=['int8', 'int16', 'int32', 'int64', 'float32', 'float64'])
    def my_row_copier_mul(x, m, y):
        input_row_count = len(x)
        for i in range(input_row_count):
            if x.is_null(i):
                y.set_null(i)
            else:
                y[i] = x[i] + x[i]
        return m * input_row_count

    # cannot overload boolean8 and int8 as boolean8 is interpreted as int8
    @omnisci('int32(Column<bool>, RowMultiplier, OutputColumn<bool>)')
    def my_row_copier_mul_bool(x, m, y):
        input_row_count = len(x)
        for i in range(input_row_count):
            if x.is_null(i):
                y.set_null(i)
            elif x[i]:
                y[i] = False
            else:
                y[i] = True
        return m * input_row_count


colnames = ['f4', 'f8', 'i1', 'i2', 'i4', 'i8', 'b']


@pytest.mark.parametrize('col', colnames)
def test_null_value(omnisci, col):
    typ = dict(f4='float32', f8='float64', i1='int8', i2='int16',
               i4='int32', i8='int64', b='bool')[col]

    if typ == 'bool':
        prefix = '_bool'
        descr, expected = omnisci.sql_execute(
            f'select {col}, not {col} from {omnisci.table_name}null')
        data, expected = zip(*list(expected))

    else:
        prefix = ''
        descr, expected = omnisci.sql_execute(
            f'select {col}, {col} + {col} from {omnisci.table_name}null')
        data, expected = zip(*list(expected))

    descr, result = omnisci.sql_execute(
        f'select * from table(my_row_copier_mul{prefix}(cursor(select {col} '
        f'from {omnisci.table_name}null), 1));')
    result, = zip(*list(result))

    assert result == expected, (result, expected, data)


def test_row_adder(omnisci):
    descr, expected = omnisci.sql_execute(
        f'select f8, f8 + f8 from {omnisci.table_name}null')
    data, expected = zip(*list(expected))

    descr, result = omnisci.sql_execute(
        'select * from table(row_adder(1, cursor(select f8, f8 '
        f'from {omnisci.table_name}null)))')
    result, = zip(*list(result))

    assert result == expected, (result, expected, data)



"""TargetInfo class specific
"""
import ctypes
import json
import warnings
from . import libfuncs
from .utils import parse_version


class TargetInfo(object):
    """Holds target device information.

    The information within a TargetInfo instance can be accessed
    only when the instance has been activated as a context manager::

      target_info = TargetInfo(<name>)
      # construct target_info
      ...

      with target_info:
          '''
          TargetInfo instance target_info is made available globally
          and it can be acquired from any Python module function, for
          example:

            from rbc.targetinfo import TargetInfo

            def foo():
                ti = TargetInfo()
                # here `ti` is the same object as `target_info`
          '''
          ...

      # target_info is now disabled globally and when a Python module
      # function tries to access it as `TargetInfo()`, a RuntimeError
      # exception will be raised.
    """

    _instance = None

    @classmethod
    def _set_instance(cls, instance):
        if instance is None:
            assert cls._instance is not None
        else:
            assert cls._instance is None
        cls._instance = instance
        return instance

    @classmethod
    def _get_instance(cls):
        if cls._instance is None:
            raise RuntimeError('Target not specified.')
        return cls._instance

    def __enter__(self):
        if self.nested:
            parent = type(self)._instance
            if parent is not None:
                type(self)._set_instance(None)
                self.update(parent)
            self._parent = parent
        return type(self)._set_instance(self)

    def __exit__(self, exc_type, exc_value, traceback):
        type(self)._set_instance(None)
        if self._parent is not None:
            assert self.nested
            type(self)._set_instance(self._parent)
        if exc_type is None:
            return True

    def __new__(cls, *args, **kwargs):
        """Return a global or a newly constructed TargetInfo instance.
        """
        if not (args or kwargs):
            return cls._get_instance()
        obj = object.__new__(cls)
        obj._init(*args, **kwargs)
        return obj

    def _init(self, name: str, strict: bool = False, nested: bool = False,
              use_tracing_allocator: bool = False):
        """
        Parameters
        ----------
        name : str
          Specify unique name of a target device.
        strict: bool
          When True, require that atomic types are concrete. Used by
          typesystem.
        nested: bool
          When True, allow nested target info contexts.
        use_tracing_allocator: bool
          When True, use the tracing allocator, to enable the LeakDetector
        """
        self.name = name
        self.strict = strict
        self.nested = nested
        self.use_tracing_allocator = use_tracing_allocator
        self._parent = None
        self.info = {}
        self.type_sizeof = {}
        self._supported_libraries = set()  # libfuncs.Library instances
        self._userdefined_externals = set()

    def __repr__(self):
        return f'{self.__class__.__name__}(name={self.name!r})'

    def add_external(self, *names):
        self._userdefined_externals.update(names)

    def add_library(self, lib):
        if isinstance(lib, str):
            lib = libfuncs.Library.get(lib)
        if isinstance(lib, libfuncs.Library):
            self._supported_libraries.add(lib)
        else:
            raise TypeError(
                f'Expected libfuncs.Library instance or library name but got {type(lib)}')

    def supports(self, name):
        """Return True if the target system defines symbol name.
        """
        if name in self._userdefined_externals:
            return True
        for lib in self._supported_libraries:
            if name in lib:
                return True
        return False

    def todict(self):
        return dict(name=self.name, strict=self.strict,
                    use_tracing_allocator=self.use_tracing_allocator,
                    info=self.info,
                    type_sizeof=self.type_sizeof,
                    libraries=[lib.name for lib in self._supported_libraries],
                    externals=list(self._userdefined_externals))

    @classmethod
    def fromdict(cls, data):
        target_info = cls(data.get('name', 'somedevice'),
                          strict=data.get('strict', False),
                          use_tracing_allocator=data.get('use_tracing_allocator', False))
        target_info.update(data)
        return target_info

    def update(self, data):
        """Update target info using other target info data. Any existing data
        bit will be overwritten except the name and strict fields.
        """
        if isinstance(data, type(self)):
            data = data.todict()
        self.info.update(data.get('info', {}))
        self.type_sizeof.update(data.get('type_sizeof', {}))
        for lib in data.get('libraries', []):
            self.add_library(lib)
        self.add_external(*data.get('externals', []))

    def tojson(self):
        return json.dumps(self.todict())

    @classmethod
    def fromjson(cls, data):
        return cls.fromdict(json.loads(data))

    @classmethod
    def dummy(cls):
        """Returns dummy target info instance.

        Warning: When a target info context already exists, its data
        is copied into the dummy target info except the name, strict
        and nested fields. The existing target info context will be
        restored when leaving the dummy target info context.
        """
        return TargetInfo(name='dummy', strict=False, nested=True)

    _host_target_info_cache = {}

    @classmethod
    def host(cls, name='host_cpu', strict=False, use_tracing_allocator=False):
        """Return target info for host CPU.
        """
        key = (name, strict, use_tracing_allocator)

        target_info = TargetInfo._host_target_info_cache.get(key)
        if target_info is not None:
            return target_info

        import llvmlite.binding as ll
        target_info = cls(name=name, strict=strict,
                          use_tracing_allocator=use_tracing_allocator)
        target_info.set('name', ll.get_host_cpu_name())
        target_info.set('triple', ll.get_default_triple())
        features = ','.join(['-+'[int(v)] + k
                             for k, v in ll.get_host_cpu_features().items()])
        target_info.set('features', features)

        for tname, ctype in dict(
                bool=ctypes.c_bool,
                size_t=ctypes.c_size_t,
                ssize_t=ctypes.c_ssize_t,
                char=ctypes.c_char,
                uchar=ctypes.c_char,
                schar=ctypes.c_char,
                byte=ctypes.c_byte,
                ubyte=ctypes.c_ubyte,
                wchar=ctypes.c_wchar,
                short=ctypes.c_short,
                ushort=ctypes.c_ushort,
                int=ctypes.c_int,
                uint=ctypes.c_uint,
                long=ctypes.c_long,
                ulong=ctypes.c_ulong,
                longlong=ctypes.c_longlong,
                ulonglong=ctypes.c_ulonglong,
                float=ctypes.c_float,
                double=ctypes.c_double,
                longdouble=ctypes.c_longdouble,
                voidptr=ctypes.c_void_p
        ).items():
            target_info.type_sizeof[tname] = ctypes.sizeof(ctype)

        target_info.add_library('m')
        target_info.add_library('stdio')
        target_info.add_library('stdlib')
        target_info.add_library('rbclib')
        if use_tracing_allocator:
            target_info.set('fn_allocate_varlen_buffer',
                            'rbclib_tracing_allocate_varlen_buffer')
            target_info.set('fn_free_buffer',
                            'rbclib_tracing_free_buffer')
        else:
            target_info.set('fn_allocate_varlen_buffer', 'rbclib_allocate_varlen_buffer')
            target_info.set('fn_free_buffer', 'rbclib_free_buffer')
        cls._host_target_info_cache[key] = target_info

        return target_info

    def set(self, prop, value):
        """Set a target device property to given value.
        """
        supported_keys = ('name', 'triple', 'datalayout', 'features', 'bits',
                          'compute_capability', 'count', 'threads', 'cores',
                          'has_cpython', 'has_numba', 'driver', 'software',
                          'llvm_version', 'null_values',
                          'fn_allocate_varlen_buffer', 'fn_free_buffer')
        if prop not in supported_keys:
            print(f'rbc.{type(self).__name__}:'
                  f' unsupported property {prop}={value}.')
        self.info[prop] = value

    # Convenience methods

    @property
    def software(self):
        """Return remote software name and version number as int tuple.
        """
        lst = self.info.get('software', '').split(None, 1)
        if len(lst) == 2:
            name, version = lst
        else:
            return lst[0], ()
        return name, parse_version(version)

    @property
    def driver(self):
        """Return device driver name and version numbers as int tuple. For
        instance::

          'CUDA', (11, 0)
        """
        lst = self.info.get('driver', '').split(None, 1)
        if len(lst) == 2:
            name, version = lst
        else:
            return lst[0], ()
        return name, parse_version(version)

    @property
    def triple(self):
        """Return target triple as a string.

        The triple is in a form "<arch>-<vendor>-<os>"
        """
        return self.info['triple']

    @property
    def arch(self):
        """Return architecture string of target device.
        """
        return self.triple.split('-', 1)[0]

    @property
    def bits(self):
        """Return target device address bit-size as int value (32, 64, ...).
        """
        bits = self.info.get('bits')
        if bits is not None:
            return bits
        # expand this dict as needed
        return dict(x86_64=64, nvptx64=64,
                    x86=32, nvptx=32, arm64=64)[self.arch]

    @property
    def datalayout(self):
        """Return LLVM datalayout of target device.
        """
        layout = self.info.get('datalayout')
        if layout is None:
            if 'cpu' not in self.name:
                print(f'rbc.{type(self).__name__}:'
                      f' no datalayout info for {self.name!r} device')
            # In the following we assume that datalayout of the target
            # host matches with the datalayout of the client host,
            # regardless of what OS these hosts run.
            layout = ''
        return layout

    @property
    def device_features(self):
        """Return a comma-separated string of CPU target features.
        """
        if 'features' in self.info:
            return ','.join(self.info['features'].split())
        return ''

    @property
    def gpu_cc(self):
        """Return compute capabilities (major, minor) of CUDA device target.
        """
        return tuple(map(int, self.info['compute_capability'].split('.')))

    @property
    def device_name(self):
        """Return the name of target device.
        """
        return self.info['name']

    @property
    def is_cpu(self):
        """Return True if the target device is CPU.
        """
        return 'cpu' in self.name

    @property
    def is_gpu(self):
        """Return True if the target device is GPU.
        """
        return 'gpu' in self.name

    @property
    def has_numba(self):
        """Check if target supports numba symbols
        """
        return self.info.get('has_numba', False)

    @property
    def has_numpy(self):
        """Check if target supports numpy symbols
        """
        return self.info.get('has_numpy', False)

    @property
    def has_cpython(self):
        """Check if target supports Python C/API symbols
        """
        return self.info.get('has_cpython', False)

    @property
    def llvm_version(self):
        """Return target system LLVM versions tuple.
        """
        return self.info.get('llvm_version')

    @property
    def null_values(self):
        """Return the null values for scalar types serialized as integers.
        """
        null_values = self.info.get('null_values', {})
        if not null_values:
            raise RuntimeError('null value support requires omniscidb-internal PR 5104')
        return null_values

    # TODO: info may also contain: count, threads, cores

    def check_enabled(self, desired_devices):
        """Check if target supports desired devices

        Checks if target is in a set of desired devices. If not, the
        target is considered disabled by the caller.

        Parameters
        ----------
        desired_devices: list
          Device names that is a subset of `['cpu', 'gpu', 'cuda']` or
          None.

        Returns
        -------
        is_contained : bool
          Return True when `desired_devices` is None or when the
          target device name is listed in `desired_devices`.
          Otherwise, return False.
        """
        if desired_devices is None:
            return True
        for d in desired_devices:
            d = d.lower()
            if d == 'cpu':
                if self.is_cpu:
                    return True
            elif d == 'gpu':
                if self.is_gpu:
                    return True
            elif d == 'cuda':
                if self.driver[0] == 'CUDA':
                    return True
            else:
                warnings.warn(
                    f'Ignoring unknown desired device specification: {d}')
        return False

    # Worker methods

    def sizeof(self, t):
        """Return the sizeof(t) value for given target device.

        Parameters
        ----------
        t : {str, ...}
            Specify types name. For complete support, one should
            implement the sizeof for the following type names: char,
            uchar, schar, byte, ubyte, short, ushort, int, uint, long,
            ulong, longlong, ulonglong, float, double, longdouble,
            complex, bool, size_t, ssize_t, wchar.

        Returns
        -------
        size : {int, None}
          Byte-size of the input type. Dummy target info will not try
          to guess the sizeof of type and will return None.
        """
        s = self.type_sizeof.get(t)
        if s is not None:
            return s
        if isinstance(t, str):
            if t == 'complex':
                return self.sizeof('float') * 2
            if t == 'double':
                return 8  # IEC 60559
            if t == 'float':
                return 4  # IEC 60559
            if t == 'char':
                return 1  # IEC 9899
        if self.name == 'dummy':
            # don't guess
            return
        if isinstance(t, str):
            if t == 'int':
                warnings.warn('Using sizeof(int) == 4')
                return 4  # this is a guess
            if t == 'long':
                warnings.warn('Using sizeof(long) == 8')
                return 8  # this is a guess
            if t == 'longlong':
                warnings.warn('Using sizeof(long long) == 8')
                return 8  # this is a guess
            if t == 'size_t':
                return self.bits // 8
        if isinstance(t, type) and issubclass(t, ctypes._SimpleCData):
            return ctypes.sizeof(t)
        raise NotImplementedError(
            f"{type(self).__name__}[{self.name},{self.triple}].sizeof({t!r})")
